# Copyright (c) Facebook, Inc. All Rights Reserved

# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/hubert/hubconf.py ]
#   Synopsis     [ the HuBERT torch hubconf ]
#   Author       [ S3PRL / Kushal Lakhotia ]
"""*********************************************************************************************"""


import logging
import os
import time
from pathlib import Path

from filelock import FileLock

from s3prl.util.download import _urls_to_filepaths

from .convert import load_and_convert_fairseq_ckpt
from .expert import LegacyUpstreamExpert as _LegacyUpstreamExpert
from .expert import UpstreamExpert as _UpstreamExpert

logger = logging.getLogger(__name__)

NEW_ENOUGH_SECS = 2.0


def hubert_custom(
    ckpt: str,
    legacy: bool = False,
    fairseq: bool = False,
    refresh: bool = False,
    **kwargs,
):
    assert not (legacy and fairseq), (
        "The option 'legacy' will directly load a fairseq checkpoint, "
        "while the option 'fairseq' will first convert the fairseq checkpoint to "
        "be fairseq indenpendent and then load the checkpoint. "
        "These two options cannot be used jointly."
    )

    if ckpt.startswith("http"):
        ckpt = _urls_to_filepaths(ckpt, refresh=refresh)

    if fairseq:
        ckpt: Path = Path(ckpt)
        converted_ckpt = ckpt.parent / f"{ckpt.stem}.converted.pt"
        lock_file = Path(str(converted_ckpt) + ".lock")

        logger.info(f"Converting a fairseq checkpoint: {ckpt}")
        logger.info(f"To: {converted_ckpt}")

        with FileLock(str(lock_file)):
            if not converted_ckpt.is_file() or (
                refresh and (time.time() - os.path.getmtime(ckpt)) > NEW_ENOUGH_SECS
            ):
                load_and_convert_fairseq_ckpt(ckpt, converted_ckpt)

        ckpt = converted_ckpt

    assert os.path.isfile(ckpt)
    if legacy:
        return _LegacyUpstreamExpert(ckpt, **kwargs)
    else:
        return _UpstreamExpert(ckpt, **kwargs)


def hubert_local(*args, **kwargs):
    return hubert_custom(*args, **kwargs)


def hubert_url(*args, **kwargs):
    return hubert_custom(*args, **kwargs)


def hubert(refresh=False, *args, **kwargs):
    """
    The default model - Base
        refresh (bool): whether to download ckpt/config again if existed
    """
    return hubert_base(refresh=refresh, *args, **kwargs)


def hubert_base(refresh=False, legacy=False, **kwargs):
    """
    The Base model
        refresh (bool): whether to download ckpt/config again if existed
    """
    kwargs["ckpt"] = "https://dl.fbaipublicfiles.com/hubert/hubert_base_ls960.pt"
    if not legacy:
        kwargs[
            "ckpt"
        ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/hubert_base_ls960.pt"
    return hubert_custom(refresh=refresh, legacy=legacy, **kwargs)


def hubert_large_ll60k(refresh=False, legacy=False, **kwargs):
    """
    The Large model
        refresh (bool): whether to download ckpt/config again if existed
    """
    kwargs["ckpt"] = "https://dl.fbaipublicfiles.com/hubert/hubert_large_ll60k.pt"
    if not legacy:
        kwargs[
            "ckpt"
        ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/hubert_large_ll60k.pt"
    return hubert_custom(refresh=refresh, legacy=legacy, **kwargs)


def hubert_base_robust_mgr(refresh=False, legacy=False, **kwargs):
    """
    The Base model, continually trained with Libri 960 hr with Musan noise, Gaussian noise and Reverberation.
        refresh (bool): whether to download ckpt/config again if existed
    """
    kwargs[
        "ckpt"
    ] = "https://huggingface.co/kphuang68/HuBERT_base_robust_mgr/resolve/main/HuBERT_base_robust_mgr_best_loss_2.7821.pt"
    if not legacy:
        kwargs[
            "ckpt"
        ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/HuBERT_base_robust_mgr_best_loss_2.7821.pt"
    return hubert_custom(refresh=refresh, legacy=legacy, **kwargs)


def mhubert_base_vp_en_es_fr_it3(refresh=False, **kwds):
    kwds[
        "ckpt"
    ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/mhubert_base_vp_en_es_fr_it3.pt"
    return hubert_custom(refresh=refresh, **kwds)


def contentvec(refresh=False, **kwds):
    kwds[
        "ckpt"
    ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/contentvec_km100.pt"
    return hubert_custom(refresh=refresh, **kwds)


def contentvec_km100(refresh=False, **kwds):
    kwds[
        "ckpt"
    ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/contentvec_km100.pt"
    return hubert_custom(refresh=refresh, **kwds)


def contentvec_km500(refresh=False, **kwds):
    kwds[
        "ckpt"
    ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/contentvec_km500.pt"
    return hubert_custom(refresh=refresh, **kwds)


def ms_hubert(refresh=False, **kwds):
    kwds["ckpt"] = "https://huggingface.co/s3prl/MS-HuBERT/resolve/main/iter3.pt"
    return hubert_custom(refresh=refresh, **kwds)



"""
Evaluation metrics
"""

from .common import accuracy, cer, compute_eer, compute_minDCF, per, ter, wer
from .diarization import calc_diarization_error
from .slot_filling import slot_edit_f1, slot_type_f1, slot_value_cer, slot_value_wer

__all__ = [
    "accuracy",
    "ter",
    "wer",
    "per",
    "cer",
    "compute_eer",
    "compute_minDCF",
    "calc_diarization_error",
    "slot_edit_f1",
    "slot_value_cer",
    "slot_value_wer",
    "slot_type_f1",
]



import logging
import os
import tempfile

import pytest
from dotenv import dotenv_values

from s3prl.dataio.corpus.librispeech import LibriSpeech
from s3prl.dataio.encoder.tokenizer import load_tokenizer
from s3prl.dataio.encoder.vocabulary import generate_vocab

SAMPLE = "GOOD MORNING MY FRIEND"


def is_same_vocab(vocabs_1, vocabs_2):
    if len(vocabs_1) != len(vocabs_2):
        return False

    for v1, v2 in zip(vocabs_1, vocabs_2):
        if v1 != v2:
            return False

    return True


@pytest.mark.corpus
def test_vocabulary():
    config = dotenv_values()
    corpus = LibriSpeech(config["LibriSpeech"])
    text_list = corpus.data_dict["train-clean-100"]["text_list"]

    with tempfile.TemporaryDirectory() as directory:
        logging.info(directory)
        text_file = os.path.join(directory, "text.txt")

        with open(text_file, "w") as fp:
            for text in text_list:
                fp.write(text + "\n")

        # Character
        char_vocabs_1 = generate_vocab("character", text_list=text_list)
        char_vocabs_2 = generate_vocab("character", text_file=text_file)

        assert isinstance(char_vocabs_1, list)
        assert isinstance(char_vocabs_2, list)
        assert is_same_vocab(char_vocabs_1, char_vocabs_2)

        char_tokenizer = load_tokenizer("character", vocab_list=char_vocabs_1)
        assert char_tokenizer.decode(char_tokenizer.encode(SAMPLE)) == SAMPLE

        # Word
        word_vocabs_1 = generate_vocab("word", text_list=text_list, vocab_size=5000)
        word_vocabs_2 = generate_vocab("word", text_file=text_file, vocab_size=5000)

        assert isinstance(word_vocabs_1, list)
        assert isinstance(word_vocabs_2, list)
        assert is_same_vocab(word_vocabs_1, word_vocabs_2)

        word_tokenizer = load_tokenizer("word", vocab_list=word_vocabs_1)
        assert word_tokenizer.decode(word_tokenizer.encode(SAMPLE)) == SAMPLE

        # Subword
        vocab_file_1 = os.path.join(directory, "subword_1")
        vocab_file_2 = os.path.join(directory, "subword_2")

        subword_vocabs_1 = generate_vocab(
            "subword", text_list=text_list, vocab_size=500, output_file=vocab_file_1
        )
        subword_vocabs_2 = generate_vocab(
            "subword", text_file=text_file, vocab_size=500, output_file=vocab_file_2
        )

        subword_tokenizer_1 = load_tokenizer(
            "subword", vocab_file=vocab_file_1 + ".model"
        )
        subword_tokenizer_2 = load_tokenizer(
            "subword", vocab_file=vocab_file_2 + ".model"
        )
        assert subword_tokenizer_1.decode(subword_tokenizer_1.encode(SAMPLE)) == SAMPLE
        assert subword_tokenizer_2.decode(subword_tokenizer_2.encode(SAMPLE)) == SAMPLE
        assert subword_tokenizer_1.encode(SAMPLE) == subword_tokenizer_2.encode(SAMPLE)



"""
Parse the QUESST14 corpus

Authors:
  * Leo 2022
  * Cheng Liang 2022
"""

import logging
import re
from pathlib import Path

logger = logging.getLogger(__name__)

__all__ = [
    "Quesst14",
]


class Quesst14:
    def __init__(self, dataset_root: str):
        dataset_root = Path(dataset_root)
        self.doc_paths = self._english_audio_paths(
            dataset_root, "language_key_utterances.lst"
        )
        self.dev_query_paths = self._english_audio_paths(
            dataset_root, f"language_key_dev.lst"
        )
        self.eval_query_paths = self._english_audio_paths(
            dataset_root, f"language_key_eval.lst"
        )

        self.n_dev_queries = len(self.dev_query_paths)
        self.n_eval_queries = len(self.eval_query_paths)
        self.n_docs = len(self.doc_paths)

    @staticmethod
    def _english_audio_paths(dataset_root_path, lst_name):
        """Extract English audio paths."""
        audio_paths = []

        with open(dataset_root_path / "scoring" / lst_name) as f:
            for line in f:
                audio_path, lang = tuple(line.strip().split())
                if lang != "nnenglish":
                    continue
                audio_path = re.sub(r"^.*?\/", "", audio_path)
                audio_paths.append(dataset_root_path / audio_path)

        return audio_paths

    @property
    def valid_queries(self):
        return self.dev_query_paths

    @property
    def test_queries(self):
        return self.eval_query_paths

    @property
    def docs(self):
        """
        Valid and Test share the same document database
        """
        return self.doc_paths

    @classmethod
    def download_dataset(cls, tgt_dir: str) -> None:
        import os
        import tarfile

        import requests

        assert os.path.exists(
            os.path.abspath(tgt_dir)
        ), "Target directory does not exist"

        def unzip_targz_then_delete(filepath: str):
            with tarfile.open(os.path.abspath(filepath)) as tar:
                tar.extractall(path=os.path.abspath(tgt_dir))
            os.remove(os.path.abspath(filepath))

        def download_from_url(url: str):
            filename = url.split("/")[-1].replace(" ", "_")
            filepath = os.path.join(tgt_dir, filename)

            r = requests.get(url, stream=True)
            if r.ok:
                logger.info(f"Saving {filename} to", os.path.abspath(filepath))
                with open(filepath, "wb") as f:
                    for chunk in r.iter_content(chunk_size=1024 * 1024 * 10):
                        if chunk:
                            f.write(chunk)
                            f.flush()
                            os.fsync(f.fileno())
                logger.info(f"{filename} successfully downloaded")
                unzip_targz_then_delete(filepath)
            else:
                logger.info(f"Download failed: status code {r.status_code}\n{r.text}")

        if not os.path.exists(
            os.path.join(os.path.abspath(tgt_dir), "quesst14Database/")
        ):
            download_from_url("https://speech.fit.vutbr.cz/files/quesst14Database.tgz")
        logger.info(
            f"Quesst14 dataset downloaded. Located at {os.path.abspath(tgt_dir)}/quesst14Database/"
        )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import logging
import math
import random
from dataclasses import dataclass, field
from typing import Dict, Optional, Tuple

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from ..roberta.roberta_model import init_bert_params
from ..wav2vec2.wav2vec2_model import (
    ChoiceEnum,
    GradMultiply,
    LayerNorm,
    MultiheadAttention,
    SamePad,
    compute_mask_indices,
    get_activation_fn,
    index_put,
)
from .load_model import MAE_AST_Pretraining_Task

logger = logging.getLogger(__name__)

MASKING_DISTRIBUTION_CHOICES = ChoiceEnum(["static", "uniform", "normal", "poisson"])


@dataclass
class MAE_AST_Config:
    # Patching settings (frame vs patch based model)
    ast_kernel_size_chan: int = field(
        default=16,
        metadata={
            "help": "When using reconstruction on image data, this sets the kernel size for channels (size and stride must be identical for reconstruction) used for masked features. Default 16, see ast_models.py"
        },
    )
    ast_kernel_size_time: int = field(
        default=16,
        metadata={
            "help": "When using reconstruction on image data, this sets the kernel size for time (size and stride must be identical for reconstruction) used for masked features. Default 16, see ast_models.py."
        },
    )
    ast_kernel_stride_chan: int = field(
        default=16,
        metadata={
            "help": "When using reconstruction on image data, this sets the kernel stride for channels (size and stride must be identical for reconstruction) used for masked features. Default 16, see ast_models.py"
        },
    )
    ast_kernel_stride_time: int = field(
        default=16,
        metadata={
            "help": "When using reconstruction on image data, this sets the kernel stride for time (size and stride must be identical for reconstruction) used for masked features. Default 16, see ast_models.py."
        },
    )

    # Encoder and general transformer settings
    encoder_layers: int = field(
        default=12, metadata={"help": "num encoder layers in the encoder transformer"}
    )
    encoder_embed_dim: int = field(
        default=768, metadata={"help": "encoder embedding dimension"}
    )
    encoder_ffn_embed_dim: int = field(
        default=3072, metadata={"help": "encoder embedding dimension for FFN"}
    )
    encoder_attention_heads: int = field(
        default=12, metadata={"help": "num encoder attention heads"}
    )
    activation_fn: str = field(
        default="gelu", metadata={"help": "activation function to use"}
    )

    layer_norm_first: bool = field(
        default=False,
        metadata={"help": "apply layernorm first in the transformer"},
    )
    feature_grad_mult: float = field(
        default=1.0,
        metadata={"help": "multiply feature extractor var grads by this"},
    )

    # Decoder settings
    use_post_enc_proj: bool = field(
        default=False,
        metadata={
            "help": "Linear projection on the encoder output. Required if decoder embed dim != encoder embed dim"
        },
    )
    decoder_embed_dim: int = field(
        default=768, metadata={"help": "decoder embedding dimension"}
    )
    decoder_layers: int = field(
        default=2, metadata={"help": "num encoder layers in the decoder transformer"}
    )
    decoder_layerdrop: float = field(
        default=0.0,
        metadata={
            "help": "probability of dropping a decoder transformer layer. The decoder is shallow, so this should typically be 0"
        },
    )

    # Dropouts
    dropout: float = field(
        default=0.1,
        metadata={"help": "dropout probability for the transformer"},
    )
    attention_dropout: float = field(
        default=0.1,
        metadata={"help": "dropout probability for attention weights"},
    )
    activation_dropout: float = field(
        default=0.0,
        metadata={"help": "dropout probability after activation in FFN"},
    )
    encoder_layerdrop: float = field(
        default=0.0,
        metadata={"help": "probability of dropping a transformer layer"},
    )
    dropout_input: float = field(
        default=0.0,
        metadata={"help": "dropout to apply to the input (after feat extr)"},
    )

    # Overall Masking Settings
    random_mask_prob: float = field(
        default=0.75,
        metadata={
            "help": "Probability of a given token being masked. Exact use depends on mask type"
        },
    )

    # Wav2Vec2-like Masking settings
    mask_length: int = field(default=10, metadata={"help": "mask length"})

    mask_selection: MASKING_DISTRIBUTION_CHOICES = field(
        default="static", metadata={"help": "how to choose mask length"}
    )
    mask_other: float = field(
        default=0,
        metadata={
            "help": "secondary mask argument "
            "(used for more complex distributions), "
            "see help in compute_mask_indicesh"
        },
    )
    no_mask_overlap: bool = field(
        default=False, metadata={"help": "whether to allow masks to overlap"}
    )
    mask_min_space: int = field(
        default=0,
        metadata={"help": "min space between spans (if no overlap is enabled)"},
    )

    # Convolutional positional embeddings (not used for MAE-AST)
    conv_pos: int = field(
        default=128,
        metadata={"help": "number of filters for convolutional positional embeddings"},
    )
    conv_pos_groups: int = field(
        default=16,
        metadata={"help": "number of groups for convolutional positional embedding"},
    )

    # loss computation
    checkpoint_activations: bool = field(
        default=False,
        metadata={"help": "recompute activations and save memory for extra compute"},
    )

    # positional embeddings
    max_token_length: int = field(
        default=48000,
        metadata={
            "help": "the longest input sequence length, used for sinusoidal positional embedding"
        },
    )
    enc_sine_pos: bool = field(
        default=False,
        metadata={"help": "sinusoidal positional embeddings for encoder input"},
    )
    enc_conv_pos: bool = field(
        default=False,
        metadata={"help": "convnet positional embeddings for encoder input"},
    )
    dec_sine_pos: bool = field(
        default=False,
        metadata={"help": "sinusoidal positional embeddings for decoder input"},
    )
    dec_conv_pos: bool = field(
        default=False,
        metadata={"help": "convnet positional embeddings for decoder input"},
    )


class MAE_AST(nn.Module):
    def __init__(
        self,
        cfg: MAE_AST_Config,
        task_cfg: MAE_AST_Pretraining_Task,
    ) -> None:
        super().__init__()
        logger.info(f"MAEModel Config: {cfg}")
        self.cfg = cfg
        self.task_cfg = task_cfg

        self.feature_extractor = nn.Identity()
        self.post_extract_proj = nn.Linear(
            cfg.ast_kernel_size_time * cfg.ast_kernel_size_chan, cfg.encoder_embed_dim
        )
        self.layer_norm = LayerNorm(task_cfg.feature_dim)
        self.batch_norm = nn.BatchNorm2d(num_features=1, affine=False)
        self.unfold = nn.Unfold(
            kernel_size=(cfg.ast_kernel_size_time, cfg.ast_kernel_size_chan),
            stride=(cfg.ast_kernel_stride_time, cfg.ast_kernel_stride_chan),
        )

        self.is_batched_mask = (
            self.task_cfg.mask_type == "random_mask_batched"
            or self.task_cfg.mask_type == "chunk_patch"
        )

        self.mask_selection = cfg.mask_selection
        self.mask_other = cfg.mask_other
        self.mask_length = cfg.mask_length
        self.no_mask_overlap = cfg.no_mask_overlap
        self.mask_min_space = cfg.mask_min_space

        self.dropout_input = nn.Dropout(cfg.dropout_input)

        self.feature_grad_mult = cfg.feature_grad_mult

        self.encoder_mask_emb = nn.Parameter(
            torch.FloatTensor(cfg.encoder_embed_dim).uniform_()
        )

        if self.cfg.enc_conv_pos:
            self.enc_conv_pos_embed = ConvPosEmbed(cfg)
        if self.cfg.enc_sine_pos:
            self.enc_sine_pos_embed = SinusoidalPositionalEncoding(
                d_model=self.cfg.encoder_embed_dim, max_len=self.cfg.max_token_length
            )

        self.encoder = TransformerEncoder(cfg)

        if self.cfg.use_post_enc_proj:
            self.post_enc_proj = nn.Linear(
                self.cfg.encoder_embed_dim, self.cfg.decoder_embed_dim
            )
        else:
            assert (
                self.cfg.decoder_embed_dim == self.cfg.encoder_embed_dim
            ), "Need post_enc_projection if encoder and decoder embed dims differ"

        self.decoder_mask_emb = nn.Parameter(
            torch.FloatTensor(cfg.decoder_embed_dim).uniform_()
        )
        if self.cfg.dec_conv_pos:
            self.dec_conv_pos_embed = ConvPosEmbed(cfg)
        if self.cfg.dec_sine_pos:
            self.dec_sine_pos_embed = SinusoidalPositionalEncoding(
                d_model=self.cfg.decoder_embed_dim, max_len=self.cfg.max_token_length
            )
        # decoder uses the same set of params as the encoders, except for the number of layers and width
        self.decoder = TransformerEncoder(cfg, decoder=True)

        self.final_proj_reconstruction = nn.Linear(
            cfg.decoder_embed_dim, cfg.ast_kernel_size_time * cfg.ast_kernel_size_chan
        )
        self.final_proj_classification = nn.Linear(
            cfg.decoder_embed_dim, cfg.ast_kernel_size_time * cfg.ast_kernel_size_chan
        )

        assert self.task_cfg.feature_type != "wav"

    def upgrade_state_dict_named(self, state_dict, name):
        """Upgrade a (possibly old) state dict for new versions of fairseq."""

        super().upgrade_state_dict_named(state_dict, name)
        return state_dict

    @classmethod
    def build_model(cls, cfg: MAE_AST_Config, task: MAE_AST_Pretraining_Task):
        """Build a new model instance."""

        model = MAE_AST(cfg, task.cfg)
        return model

    def forward_features(self, source: torch.Tensor) -> torch.Tensor:
        if self.feature_grad_mult > 0:
            features = self.feature_extractor(source)
            if self.feature_grad_mult != 1.0:
                features = GradMultiply.apply(features, self.feature_grad_mult)
        else:
            with torch.no_grad():
                features = self.feature_extractor(source)
        return features

    def forward_padding_mask(
        self,
        features: torch.Tensor,
        padding_mask: torch.Tensor,
        feature_dim=128,
    ) -> torch.Tensor:
        if padding_mask[:, -1].sum() == 0:  # Fast exit during training if not necessary
            return padding_mask.new_zeros(features.shape[:2])

        non_zero_count = padding_mask.size(-1) - padding_mask.sum(dim=-1)
        num_patches_over_channel = feature_dim // self.cfg.ast_kernel_size_chan
        padding_mask_indices = (
            ((non_zero_count - 1) // self.cfg.ast_kernel_size_time) + 1
        ) * num_patches_over_channel
        new_padding_mask = padding_mask.new_zeros(features.shape[:2])

        for i in range(new_padding_mask.size(0)):
            if padding_mask_indices[i] < new_padding_mask.size(-1):
                new_padding_mask[i, padding_mask_indices[i] :] = True

        return new_padding_mask

    def forward_mask(self, features, padding_mask):
        B, T, C = features.shape
        num_retained_tokens = int((1 - self.cfg.random_mask_prob) * T)
        num_retained_tokens = max(1, num_retained_tokens)
        retained_idx = []
        masked_idx = []

        if self.task_cfg.mask_type == "retain_spans":
            num_retained_tokens = 0
            while (
                num_retained_tokens == 0
            ):  # This loop will almost never run more than once for any reasonable mask ratio.
                retained_indices = compute_mask_indices(
                    (B, T),
                    padding_mask,
                    self.cfg.random_mask_prob,
                    self.mask_length,
                    self.mask_selection,
                    self.mask_other,
                    min_masks=2,
                    no_overlap=self.no_mask_overlap,
                    min_space=self.mask_min_space,
                )
                num_retained_tokens = retained_indices[0].sum()
            for i in range(B):
                cur_retained_idx = np.where(retained_indices[i])[0]
                cur_masked_idx = np.where(~retained_indices[i])[0]
                retained_idx.append(cur_retained_idx)
                features[i, cur_masked_idx] = self.encoder_mask_emb
                masked_idx.append(cur_masked_idx)
        elif self.task_cfg.mask_type == "random_mask":
            for i in range(B):
                idx = list(range(T))
                random.shuffle(idx)
                cur_retained_idx = idx[:num_retained_tokens]
                retained_idx.append(cur_retained_idx)
                cur_masked_idx = idx[num_retained_tokens:]
                masked_idx.append(cur_masked_idx)
                features[i, cur_masked_idx] = self.encoder_mask_emb
        elif self.task_cfg.mask_type == "random_mask_batched":
            idx = list(range(T))
            random.shuffle(idx)
            cur_retained_idx = idx[:num_retained_tokens]
            retained_idx = [cur_retained_idx]
            cur_masked_idx = idx[num_retained_tokens:]
            masked_idx = [cur_masked_idx]
            features[:, cur_masked_idx] = self.encoder_mask_emb
        elif (
            self.task_cfg.mask_type == "chunk_mask"
        ):  # Copies SSAST code and goes to bottom right from uniform starting index
            cur_masked_idx = set()
            chunk_size = random.randrange(3, 5 + 1)
            chan_adjust = self.task_cfg.feature_dim // self.cfg.ast_kernel_stride_chan
            num_masked_tokens = T - num_retained_tokens
            while len(cur_masked_idx) <= num_masked_tokens:
                t_topleft = random.randrange(T)
                for t_offset in range(0, chunk_size):
                    for c_offset in range(0, chunk_size):
                        mask_cand = t_topleft + t_offset + chan_adjust * c_offset
                        if mask_cand < T:
                            cur_masked_idx.add(mask_cand)
            cur_masked_idx = list(cur_masked_idx)
            cur_masked_idx = cur_masked_idx[:num_masked_tokens]
            cur_retained_idx = list(set(range(T)).difference(cur_masked_idx))
            for i in range(
                B
            ):  # Using same mask for whole batch because SSAST code is very slow
                retained_idx.append(cur_retained_idx)
                masked_idx.append(cur_masked_idx)
                features[i, cur_masked_idx] = self.encoder_mask_emb

        return retained_idx, masked_idx, T - num_retained_tokens

    def forward(
        self,
        source: torch.Tensor,
        padding_mask: Optional[torch.Tensor] = None,
        mask: bool = True,
        features_only: bool = False,
        output_layer: Optional[int] = None,
        is_decoder_finetune: bool = False,
        is_input_prepatched: bool = False,
    ) -> Dict[str, torch.Tensor]:
        # Checks whether the dataset was patched and normalized before-hand. is_input_prepatched == True for speed profiling during training.
        if is_input_prepatched:
            source_patch = source
        else:
            # Batch normalization ('Mimics' AST dataset normalization)
            source = source.unsqueeze(1)
            source = self.batch_norm(source) * 0.5  # Mean 0, St dev 0.5
            # Create image patches for masking via Unfold. BTC input shape BTC output shape
            # Output continues to be unsqueezed from batch norm
            source_patch = self.unfold(source).transpose(-1, -2)

        features = self.forward_features(source_patch)
        if padding_mask is not None:  # Reshape padding mask
            padding_mask = self.forward_padding_mask(features, padding_mask)
        if (
            self.post_extract_proj is not None
        ):  # Project patches to vectors of size encoder_dim
            features = self.post_extract_proj(features)
        if mask:  # additional regularization adopted from hubert
            features = self.dropout_input(features)

        B, T, C = features.shape

        # Calculate retained_idx and masked_idx. Uses safe assumption that nothing is padded during pretraining
        if mask:
            retained_idx, masked_idx, num_masked_tokens = self.forward_mask(
                features, padding_mask
            )
        else:
            retained_idx = []
            masked_idx = []
            num_masked_tokens = 0

        # Pre-Encoder Positional Embeddings
        if self.cfg.enc_conv_pos:
            conv_pos = self.enc_conv_pos_embed(features, padding_mask)
            features = conv_pos + features
        if self.cfg.enc_sine_pos:
            sine_pos = self.enc_sine_pos_embed(features, padding_mask)
            features = sine_pos + features

        # Remove masked tokens from features
        if mask:
            if self.is_batched_mask:
                x = features[:, retained_idx[0]]
                retained_padding_mask = padding_mask[:, retained_idx[0]]
            else:
                x = []
                retained_padding_mask = []
                for i in range(B):
                    x.append(features[i, retained_idx[i]])
                    retained_padding_mask.append(padding_mask[i, retained_idx[i]])
                x = torch.stack(x, dim=0)
                retained_padding_mask = torch.stack(retained_padding_mask, dim=0)
        else:
            x = features
            retained_padding_mask = padding_mask

        # Encoder forward pass + Early return for features
        x, encoder_hidden_states = self.encoder(
            x,
            padding_mask=retained_padding_mask,
            layer=None if output_layer is None else output_layer - 1,
        )

        if not is_decoder_finetune and (features_only or not mask):
            return {
                "x": x,
                "padding_mask": retained_padding_mask,
                "features": features,
                "hidden_states": encoder_hidden_states,
            }

        if self.cfg.use_post_enc_proj:
            x = self.post_enc_proj(x)

        # Add masked tokens back
        if mask:
            full_x = torch.empty((B, T, C), device=x.device, dtype=x.dtype)
            mask_indices = torch.zeros(
                torch.Size([B, T]), device=padding_mask.device, dtype=torch.bool
            )
            if self.is_batched_mask:
                full_x[:, retained_idx[0]] = x
                full_x[:, masked_idx[0]] = self.decoder_mask_emb
                mask_indices[:, masked_idx[0]] = True
            else:
                for i, (cur_feat, ridx, midx) in enumerate(
                    zip(x, retained_idx, masked_idx)
                ):
                    full_x[i, ridx] = cur_feat
                    full_x[i, midx] = self.decoder_mask_emb
                    mask_indices[i, midx] = True
        else:
            full_x = x

        # Pre decoder positional embeddings
        if self.cfg.dec_conv_pos:
            conv_pos = self.dec_conv_pos_embed(full_x, padding_mask)
            full_x = conv_pos + full_x
        if self.cfg.dec_sine_pos:
            # Concerning that magnitudes of layer-normed encoder outputs are similar to decoder positional embeddings.
            sine_pos = self.dec_sine_pos_embed(full_x, padding_mask)
            full_x = sine_pos + full_x

        # Decoder forward pass
        x, decoder_hidden_states = self.decoder(
            full_x, padding_mask=padding_mask, layer=None
        )

        if is_decoder_finetune:
            return {
                "x": x,
                "padding_mask": padding_mask,
                "features": features,
                "hidden_states": encoder_hidden_states + decoder_hidden_states,
            }

        # Construct linear projection logits and masked reconstruction targets
        x_masked_indices = x[mask_indices].view(B, num_masked_tokens, -1)

        logit_m_list_recon = self.final_proj_reconstruction(x_masked_indices)
        logit_m_list_class = self.final_proj_classification(x_masked_indices)

        target_m_list = source_patch[mask_indices].view(B, num_masked_tokens, -1)

        result = {
            "logit_m_list_recon": logit_m_list_recon,
            "logit_m_list_class": logit_m_list_class,
            "target_m_list": target_m_list,
            "padding_mask": padding_mask,
        }
        return result

    def extract_features(
        self,
        source: torch.Tensor,
        padding_mask: Optional[torch.Tensor] = None,
        mask: bool = False,
        ret_conv: bool = False,
        output_layer: Optional[int] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        res = self.forward(
            source,
            padding_mask=padding_mask,
            mask=mask,
            features_only=True,
            output_layer=output_layer,
        )
        feature = res["features"] if ret_conv else res["x"]
        return feature, res["padding_mask"]

    def get_logits(self, net_output):
        logits_list = net_output["logit_m_list_recon"], net_output["logit_m_list_class"]
        logits_list = [x.float() for x in logits_list if x is not None]
        return logits_list

    def get_targets(self, net_output, is_masked=True):
        return net_output["target_m_list"].float()

    def get_extra_losses(self, net_output):
        extra_losses = []
        names = []

        if "features_pen" in net_output:
            extra_losses.append(net_output["features_pen"])
            names.append("features_pen")

        return extra_losses, names

    def remove_pretraining_modules(self):
        self.final_proj_reconstruction = None
        self.final_proj_classification = None


class ConvPosEmbed(nn.Module):
    def __init__(self, args, decoder=False):
        super().__init__()
        self.embedding_dim = args.encoder_embed_dim
        self.pos_conv = nn.Conv1d(
            self.embedding_dim,
            self.embedding_dim,
            kernel_size=args.conv_pos,
            padding=args.conv_pos // 2,
            groups=args.conv_pos_groups,
        )
        dropout = 0
        std = math.sqrt((4 * (1.0 - dropout)) / (args.conv_pos * self.embedding_dim))
        nn.init.normal_(self.pos_conv.weight, mean=0, std=std)
        nn.init.constant_(self.pos_conv.bias, 0)

        self.pos_conv = nn.utils.weight_norm(self.pos_conv, name="weight", dim=2)
        self.pos_conv = nn.Sequential(self.pos_conv, SamePad(args.conv_pos), nn.GELU())

    def forward(self, x, padding_mask=None):
        if padding_mask is not None:
            x = index_put(x, padding_mask, 0)

        x_conv = self.pos_conv(x.transpose(1, 2))  # [b,t,c] -> [b,c,t]
        x_conv = x_conv.transpose(1, 2)  # [b,c,t] -> [b,t,c]
        return x_conv


class TransformerEncoder(nn.Module):
    def __init__(self, args, decoder=False):
        super().__init__()

        self.dropout = args.dropout

        layers = []
        if decoder:
            # the onnly difference between encoder and decoder is the number of layers
            num_layers = args.decoder_layers
            self.embedding_dim = args.decoder_embed_dim
            self.layerdrop = args.decoder_layerdrop
        else:
            num_layers = args.encoder_layers
            self.embedding_dim = args.encoder_embed_dim
            self.layerdrop = args.encoder_layerdrop
        for _ in range(num_layers):
            layer = TransformerSentenceEncoderLayer(
                embedding_dim=self.embedding_dim,
                ffn_embedding_dim=args.encoder_ffn_embed_dim,
                num_attention_heads=args.encoder_attention_heads,
                dropout=self.dropout,
                attention_dropout=args.attention_dropout,
                activation_dropout=args.activation_dropout,
                activation_fn=args.activation_fn,
                layer_norm_first=args.layer_norm_first,
            )
            if args.checkpoint_activations:
                raise NotImplementedError(
                    "This part is not used thus not included in S3PRL, please see the official repo."
                )
            layers.append(layer)
        self.layers = nn.ModuleList(layers)

        self.layer_norm_first = args.layer_norm_first
        self.layer_norm = LayerNorm(self.embedding_dim)

        self.apply(init_bert_params)

    def forward(self, x, padding_mask=None, layer=None):
        # print(f"shape of input the transformer: {x.shape}")
        # print(f"shape of padding mask: {padding_mask.shape}")
        x, layer_results = self.extract_features(x, padding_mask, layer)

        if self.layer_norm_first and layer is None:
            x = self.layer_norm(x)

        return x, layer_results

    def extract_features(self, x, padding_mask=None, tgt_layer=None):
        if not self.layer_norm_first:
            x = self.layer_norm(x)

        x = F.dropout(x, p=self.dropout, training=self.training)

        # B x T x C -> T x B x C
        x = x.transpose(0, 1)

        hidden_states = []
        r = None
        for i, layer in enumerate(self.layers):
            dropout_probability = np.random.random()
            if not self.training or (dropout_probability > self.layerdrop):
                x, z = layer(x, self_attn_padding_mask=padding_mask, need_weights=False)
                hidden_states.append(x.transpose(0, 1))
            if i == tgt_layer:
                r = x
                break

        if r is not None:
            x = r

        # T x B x C -> B x T x C
        x = x.transpose(0, 1)

        return x, hidden_states

    def max_positions(self):
        """Maximum output length supported by the encoder."""
        return self.cfg.max_positions

    def upgrade_state_dict_named(self, state_dict, name):
        """Upgrade a (possibly old) state dict for new versions of fairseq."""
        return state_dict


class TransformerSentenceEncoderLayer(nn.Module):
    """
    Implements a Transformer Encoder Layer used in BERT/XLM style pre-trained
    models.
    """

    def __init__(
        self,
        embedding_dim: float = 768,
        ffn_embedding_dim: float = 3072,
        num_attention_heads: float = 8,
        dropout: float = 0.1,
        attention_dropout: float = 0.1,
        activation_dropout: float = 0.1,
        activation_fn: str = "relu",
        layer_norm_first: bool = False,
    ) -> None:
        super().__init__()
        # Initialize parameters
        self.embedding_dim = embedding_dim
        self.dropout = dropout
        self.activation_dropout = activation_dropout

        # Initialize blocks
        self.activation_fn = get_activation_fn(activation_fn)
        self.self_attn = MultiheadAttention(
            self.embedding_dim,
            num_attention_heads,
            dropout=attention_dropout,
            self_attention=True,
        )

        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(self.activation_dropout)
        self.dropout3 = nn.Dropout(dropout)

        self.layer_norm_first = layer_norm_first

        # layer norm associated with the self attention layer
        self.self_attn_layer_norm = LayerNorm(self.embedding_dim)
        self.fc1 = nn.Linear(self.embedding_dim, ffn_embedding_dim)
        self.fc2 = nn.Linear(ffn_embedding_dim, self.embedding_dim)

        # layer norm associated with the position wise feed-forward NN
        self.final_layer_norm = LayerNorm(self.embedding_dim)

    def forward(
        self,
        x: torch.Tensor,
        self_attn_mask: torch.Tensor = None,
        self_attn_padding_mask: torch.Tensor = None,
        need_weights: bool = False,
        att_args=None,
    ):
        """
        LayerNorm is applied either before or after the self-attention/ffn
        modules similar to the original Transformer imlementation.
        """
        residual = x

        if self.layer_norm_first:
            x = self.self_attn_layer_norm(x)
            x, attn = self.self_attn(
                query=x,
                key=x,
                value=x,
                key_padding_mask=self_attn_padding_mask,
                attn_mask=self_attn_mask,
            )
            x = self.dropout1(x)
            x = residual + x

            residual = x
            x = self.final_layer_norm(x)
            x = self.activation_fn(self.fc1(x))
            x = self.dropout2(x)
            x = self.fc2(x)
            x = self.dropout3(x)
            x = residual + x
        else:
            x, attn = self.self_attn(
                query=x,
                key=x,
                value=x,
                key_padding_mask=self_attn_padding_mask,
            )

            x = self.dropout1(x)
            x = residual + x

            x = self.self_attn_layer_norm(x)

            residual = x
            x = self.activation_fn(self.fc1(x))
            x = self.dropout2(x)
            x = self.fc2(x)
            x = self.dropout3(x)
            x = residual + x
            x = self.final_layer_norm(x)

        return x, attn


class SinusoidalPositionalEncoding(nn.Module):
    def __init__(self, d_model: int, max_len: int = 480000):
        super().__init__()
        position = torch.arange(max_len).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, d_model, 2) * (-math.log(10000.0) / d_model)
        )
        pe = torch.zeros(1, max_len, d_model)
        pe[0, :, 0::2] = torch.sin(position * div_term)
        pe[0, :, 1::2] = torch.cos(position * div_term)
        self.register_buffer("pe", pe)

    def forward(self, x, padding_mask):
        """
        Args:
            x: Tensor, shape [bsz, seq_len, embedding_dim]
        """
        pe = self.pe[:, : x.shape[1]].repeat((padding_mask.shape[0], 1, 1))
        pe[padding_mask] = 0.0
        return pe



"""
Commonly used metrics

Authors
  * Leo 2022
  * Heng-Jui Chang 2022
  * Haibin Wu 2022
"""

from typing import List, Union

import editdistance as ed
from scipy.interpolate import interp1d
from scipy.optimize import brentq
from sklearn.metrics import accuracy_score, roc_curve

__all__ = [
    "accuracy",
    "ter",
    "wer",
    "per",
    "cer",
    "compute_eer",
    "compute_minDCF",
]


def accuracy(xs, ys, item_same_fn=None):
    if isinstance(xs, (tuple, list)):
        assert isinstance(ys, (tuple, list))
        return _accuracy_impl(xs, ys, item_same_fn)
    elif isinstance(xs, dict):
        assert isinstance(ys, dict)
        keys = sorted(list(xs.keys()))
        xs = [xs[k] for k in keys]
        ys = [ys[k] for k in keys]
        return _accuracy_impl(xs, ys, item_same_fn)
    else:
        raise ValueError


def _accuracy_impl(xs, ys, item_same_fn=None):
    item_same_fn = item_same_fn or (lambda x, y: x == y)
    same = [int(item_same_fn(x, y)) for x, y in zip(xs, ys)]
    return sum(same) / len(same)


def ter(hyps: List[Union[str, List[str]]], refs: List[Union[str, List[str]]]) -> float:
    """Token error rate calculator.

    Args:
        hyps (List[Union[str, List[str]]]): List of hypotheses.
        refs (List[Union[str, List[str]]]): List of references.

    Returns:
        float: Averaged token error rate overall utterances.
    """
    error_tokens = 0
    total_tokens = 0
    for h, r in zip(hyps, refs):
        error_tokens += ed.eval(h, r)
        total_tokens += len(r)
    return float(error_tokens) / float(total_tokens)


def wer(hyps: List[str], refs: List[str]) -> float:
    """Word error rate calculator.

    Args:
        hyps (List[str]): List of hypotheses.
        refs (List[str]): List of references.

    Returns:
        float: Averaged word error rate overall utterances.
    """
    hyps = [h.split(" ") for h in hyps]
    refs = [r.split(" ") for r in refs]
    return ter(hyps, refs)


def per(hyps: List[str], refs: List[str]) -> float:
    """Phoneme error rate calculator.

    Args:
        hyps (List[str]): List of hypotheses.
        refs (List[str]): List of references.

    Returns:
        float: Averaged phoneme error rate overall utterances.
    """
    return wer(hyps, refs)


def cer(hyps: List[str], refs: List[str]) -> float:
    """Character error rate calculator.

    Args:
        hyps (List[str]): List of hypotheses.
        refs (List[str]): List of references.

    Returns:
        float: Averaged character error rate overall utterances.
    """
    return ter(hyps, refs)


def compute_eer(labels: List[int], scores: List[float]):
    """Compute equal error rate.

    Args:
        scores (List[float]): List of hypotheses.
        labels (List[int]): List of references.

    Returns:
        eer (float): Equal error rate.
        treshold (float): The treshold to accept a target trial.
    """
    fpr, tpr, thresholds = roc_curve(labels, scores, pos_label=1)
    eer = brentq(lambda x: 1.0 - x - interp1d(fpr, tpr)(x), 0.0, 1.0)
    threshold = interp1d(fpr, thresholds)(eer)
    return eer, threshold


def compute_minDCF(
    labels: List[int],
    scores: List[float],
    p_target: float = 0.01,
    c_miss: int = 1,
    c_fa: int = 1,
):
    """Compute MinDCF.
    Computes the minimum of the detection cost function.  The comments refer to
    equations in Section 3 of the NIST 2016 Speaker Recognition Evaluation Plan.

    Args:
        scores (List[float]): List of hypotheses.
        labels (List[int]): List of references.
        p (float): The prior probability of positive class.
        c_miss (int): The cost of miss.
        c_fa (int): The cost of false alarm.

    Returns:
        min_dcf (float): The calculated min_dcf.
        min_c_det_threshold (float): The treshold to calculate min_dcf.
    """
    fpr, tpr, thresholds = roc_curve(labels, scores, pos_label=1)
    fnr = 1.0 - tpr

    min_c_det = float("inf")
    min_c_det_threshold = thresholds[0]
    for i in range(0, len(fnr)):
        c_det = c_miss * fnr[i] * p_target + c_fa * fpr[i] * (1 - p_target)
        if c_det < min_c_det:
            min_c_det = c_det
            min_c_det_threshold = thresholds[i]
    c_def = min(c_miss * p_target, c_fa * (1 - p_target))
    min_dcf = min_c_det / c_def
    return min_dcf, min_c_det_threshold



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/mae_ast/hubconf.py ]
#   Synopsis     [ the MAE-AST torch hubconf ]
#   Author       [ S3PRL ]
#   Copyright    [ Copyleft(c), Speech Lab, NTU, Taiwan ]
"""*********************************************************************************************"""


import os

from s3prl.util.download import _urls_to_filepaths

from .expert import UpstreamExpert as _UpstreamExpert


def mae_ast_local(ckpt, *args, **kwargs):
    """
    The model from local ckpt
        ckpt (str): PATH
    """
    assert os.path.isfile(ckpt)
    return _UpstreamExpert(ckpt, *args, **kwargs)


def mae_ast_url(ckpt, refresh=False, *args, **kwargs):
    """
    The model from URL
        ckpt (str): URL
    """
    return mae_ast_local(_urls_to_filepaths(ckpt, refresh=refresh), *args, **kwargs)


def mae_ast_frame(refresh=False, *args, **kwargs):
    """
    The MAE-AST Frame model, 12-layered, random masking
        refresh (bool): whether to download ckpt/config again if existed
    """
    kwargs[
        "ckpt"
    ] = "https://www.cs.utexas.edu/~harwath/model_checkpoints/mae_ast/random_frame_75_12LayerEncoder.pt"
    return mae_ast_url(refresh=refresh, *args, **kwargs)


def mae_ast_patch(refresh=False, *args, **kwargs):
    """
    The MAE-AST Patch model, 12-layered, chunked masking
        refresh (bool): whether to download ckpt/config again if existed
    """
    kwargs[
        "ckpt"
    ] = "https://www.cs.utexas.edu/~harwath/model_checkpoints/mae_ast/chunk_patch_75_12LayerEncoder.pt"
    return mae_ast_url(refresh=refresh, *args, **kwargs)



"""
Metrics for the slot filling SLU task

Authors:
  * Yung-Sung Chuang 2021
  * Heng-Jui Chang 2022
"""

import re
from typing import Dict, List, Tuple

from .common import cer, wer

__all__ = ["slot_type_f1", "slot_value_cer", "slot_value_wer", "slot_edit_f1"]


def clean(ref: str) -> str:
    ref = re.sub(r"B\-(\S+) ", "", ref)
    ref = re.sub(r" E\-(\S+)", "", ref)
    return ref


def parse(hyp: str, ref: str) -> Tuple[str, str, str, str]:
    gex = re.compile(r"B\-(\S+) (.+?) E\-\1")

    hyp = re.sub(r" +", " ", hyp)
    ref = re.sub(r" +", " ", ref)

    hyp_slots = gex.findall(hyp)
    ref_slots = gex.findall(ref)

    ref_slots = ";".join([":".join([x[1], x[0]]) for x in ref_slots])
    if len(hyp_slots) > 0:
        hyp_slots = ";".join([":".join([clean(x[1]), x[0]]) for x in hyp_slots])
    else:
        hyp_slots = ""

    ref = clean(ref)
    hyp = clean(hyp)

    return ref, hyp, ref_slots, hyp_slots


def get_slot_dict(
    hyp: str, ref: str
) -> Tuple[Dict[str, List[str]], Dict[str, List[str]]]:
    ref_text, hyp_text, ref_slots, hyp_slots = parse(hyp, ref)

    ref_slots = ref_slots.split(";")
    hyp_slots = hyp_slots.split(";")
    ref_dict, hyp_dict = {}, {}

    if ref_slots[0] != "":
        for ref_slot in ref_slots:
            v, k = ref_slot.split(":")
            ref_dict.setdefault(k, [])
            ref_dict[k].append(v)

    if hyp_slots[0] != "":
        for hyp_slot in hyp_slots:
            v, k = hyp_slot.split(":")
            hyp_dict.setdefault(k, [])
            hyp_dict[k].append(v)

    return ref_dict, hyp_dict


def slot_type_f1(hypothesis: List[str], groundtruth: List[str], **kwargs) -> float:
    F1s = []
    for p, t in zip(hypothesis, groundtruth):
        ref_dict, hyp_dict = get_slot_dict(p, t)

        # Slot Type F1 evaluation
        if len(hyp_dict.keys()) == 0 and len(ref_dict.keys()) == 0:
            F1 = 1.0
        elif len(hyp_dict.keys()) == 0:
            F1 = 0.0
        elif len(ref_dict.keys()) == 0:
            F1 = 0.0
        else:
            P, R = 0.0, 0.0
            for slot in ref_dict:
                if slot in hyp_dict:
                    R += 1
            R = R / len(ref_dict.keys())
            for slot in hyp_dict:
                if slot in ref_dict:
                    P += 1
            P = P / len(hyp_dict.keys())
            F1 = 2 * P * R / (P + R) if (P + R) > 0 else 0.0
        F1s.append(F1)

    return sum(F1s) / len(F1s)


def slot_value_cer(hypothesis: List[str], groundtruth: List[str], **kwargs) -> float:
    value_hyps, value_refs = [], []
    for p, t in zip(hypothesis, groundtruth):
        ref_dict, hyp_dict = get_slot_dict(p, t)

        # Slot Value WER/CER evaluation
        unique_slots = list(ref_dict.keys())
        for slot in unique_slots:
            for ref_i, ref_v in enumerate(ref_dict[slot]):
                if slot not in hyp_dict:
                    hyp_v = ""
                    value_refs.append(ref_v)
                    value_hyps.append(hyp_v)
                else:
                    min_cer = 100
                    best_hyp_v = ""
                    for hyp_v in hyp_dict[slot]:
                        tmp_cer = cer([hyp_v], [ref_v])
                        if min_cer > tmp_cer:
                            min_cer = tmp_cer
                            best_hyp_v = hyp_v
                    value_refs.append(ref_v)
                    value_hyps.append(best_hyp_v)

    return cer(value_hyps, value_refs)


def slot_value_wer(hypothesis: List[str], groundtruth: List[str], **kwargs) -> float:
    value_hyps = []
    value_refs = []
    for p, t in zip(hypothesis, groundtruth):
        ref_dict, hyp_dict = get_slot_dict(p, t)

        # Slot Value WER/CER evaluation
        unique_slots = list(ref_dict.keys())
        for slot in unique_slots:
            for ref_i, ref_v in enumerate(ref_dict[slot]):
                if slot not in hyp_dict:
                    hyp_v = ""
                    value_refs.append(ref_v)
                    value_hyps.append(hyp_v)
                else:
                    min_wer = 100
                    best_hyp_v = ""
                    for hyp_v in hyp_dict[slot]:
                        tmp_wer = wer([hyp_v], [ref_v])
                        if min_wer > tmp_wer:
                            min_wer = tmp_wer
                            best_hyp_v = hyp_v
                    value_refs.append(ref_v)
                    value_hyps.append(best_hyp_v)

    return wer(value_hyps, value_refs)


def slot_edit_f1(
    hypothesis: List[str], groundtruth: List[str], loop_over_all_slot: bool, **kwargs
) -> float:
    slot2F1 = {}  # defaultdict(lambda: [0,0,0]) # TPs, FNs, FPs
    for p, t in zip(hypothesis, groundtruth):
        ref_dict, hyp_dict = get_slot_dict(p, t)

        # Collecting unique slots
        unique_slots = list(ref_dict.keys())
        if loop_over_all_slot:
            unique_slots += [x for x in hyp_dict if x not in ref_dict]
        # Evaluating slot edit F1
        for slot in unique_slots:
            TP = 0
            FP = 0
            FN = 0
            if slot not in ref_dict:  # this never happens in list(ref_dict.keys())
                for hyp_v in hyp_dict[slot]:
                    FP += 1
            else:
                for ref_i, ref_v in enumerate(ref_dict[slot]):
                    if slot not in hyp_dict:
                        FN += 1
                    else:
                        match = False
                        for hyp_v in hyp_dict[slot]:
                            # if ref_i < len(hyp_dict[slot]):
                            #    hyp_v = hyp_dict[slot][ref_i]
                            if hyp_v == ref_v:
                                match = True
                                break
                        if match:
                            TP += 1
                        else:
                            FN += 1
                            FP += 1
            slot2F1.setdefault(slot, [0, 0, 0])
            slot2F1[slot][0] += TP
            slot2F1[slot][1] += FN
            slot2F1[slot][2] += FP

    all_TPs, all_FNs, all_FPs = 0, 0, 0
    for slot in slot2F1.keys():
        all_TPs += slot2F1[slot][0]
        all_FNs += slot2F1[slot][1]
        all_FPs += slot2F1[slot][2]

    return 2 * all_TPs / (2 * all_TPs + all_FPs + all_FNs)


def slot_edit_f1_full(hypothesis: List[str], groundtruth: List[str], **kwargs) -> float:
    return slot_edit_f1(hypothesis, groundtruth, loop_over_all_slot=True, **kwargs)


def slot_edit_f1_part(hypothesis: List[str], groundtruth: List[str], **kwargs) -> float:
    return slot_edit_f1(hypothesis, groundtruth, loop_over_all_slot=False, **kwargs)



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
# 	The S3PRL team adopted it from Harri Taylor's implementation.
"""*********************************************************************************************"""
#   FileName     [ upstream/vggish/vggish.py ]
#   Synopsis     [ VGGish model definition ]
#   Author       [ Harri Taylor, S3PRL ]
#   Copyright    [ Copyleft(c), Harri Taylor ]
"""*********************************************************************************************"""


import torch
import torch.nn as nn

from . import vggish_params


class VGG(nn.Module):
    def __init__(self, features):
        super(VGG, self).__init__()
        self.features = features
        self.embeddings = nn.Sequential(
            nn.Linear(512 * 4 * 6, 4096),
            nn.ReLU(True),
            nn.Linear(4096, 4096),
            nn.ReLU(True),
            nn.Linear(4096, 128),
            nn.ReLU(True),
        )

    def forward(self, x):
        x = self.features(x)

        # Transpose the output from features to
        # remain compatible with vggish embeddings
        x = torch.transpose(x, 1, 3)
        x = torch.transpose(x, 1, 2)
        x = x.contiguous()
        x = x.view(x.size(0), -1)

        return self.embeddings(x)


class Postprocessor(nn.Module):
    """
    Post-processes VGGish embeddings. Returns a torch.Tensor instead of a
    numpy array in order to preserve the gradient.
    "The initial release of AudioSet included 128-D VGGish embeddings for each
    segment of AudioSet. These released embeddings were produced by applying
    a PCA transformation (technically, a whitening transform is included as well)
    and 8-bit quantization to the raw embedding output from VGGish, in order to
    stay compatible with the YouTube-8M project which provides visual embeddings
    in the same format for a large set of YouTube videos. This class implements
    the same PCA (with whitening) and quantization transformations."
    """

    def __init__(self):
        """Constructs a postprocessor."""
        super(Postprocessor, self).__init__()
        # Create empty matrix, for user"s state_dict to load
        self.pca_eigen_vectors = torch.empty(
            (
                vggish_params.EMBEDDING_SIZE,
                vggish_params.EMBEDDING_SIZE,
            ),
            dtype=torch.float,
        )
        self.pca_means = torch.empty(
            (vggish_params.EMBEDDING_SIZE, 1), dtype=torch.float
        )

        self.pca_eigen_vectors = nn.Parameter(
            self.pca_eigen_vectors, requires_grad=False
        )
        self.pca_means = nn.Parameter(self.pca_means, requires_grad=False)

    def postprocess(self, embeddings_batch):
        """Applies tensor postprocessing to a batch of embeddings.
        Args:
          embeddings_batch: An tensor of shape [batch_size, embedding_size]
            containing output from the embedding layer of VGGish.
        Returns:
          A tensor of the same shape as the input, containing the PCA-transformed,
          quantized, and clipped version of the input.
        """
        assert len(embeddings_batch.shape) == 2, "Expected 2-d batch, got %r" % (
            embeddings_batch.shape,
        )
        assert (
            embeddings_batch.shape[1] == vggish_params.EMBEDDING_SIZE
        ), "Bad batch shape: %r" % (embeddings_batch.shape,)

        # Apply PCA.
        # - Embeddings come in as [batch_size, embedding_size].
        # - Transpose to [embedding_size, batch_size].
        # - Subtract pca_means column vector from each column.
        # - Premultiply by PCA matrix of shape [output_dims, input_dims]
        #   where both are are equal to embedding_size in our case.
        # - Transpose result back to [batch_size, embedding_size].
        pca_applied = torch.mm(
            self.pca_eigen_vectors, (embeddings_batch.t() - self.pca_means)
        ).t()

        # Quantize by:
        # - clipping to [min, max] range
        clipped_embeddings = torch.clamp(
            pca_applied, vggish_params.QUANTIZE_MIN_VAL, vggish_params.QUANTIZE_MAX_VAL
        )
        # - convert to 8-bit in range [0.0, 255.0]
        quantized_embeddings = torch.round(
            (clipped_embeddings - vggish_params.QUANTIZE_MIN_VAL)
            * (
                255.0
                / (vggish_params.QUANTIZE_MAX_VAL - vggish_params.QUANTIZE_MIN_VAL)
            )
        )
        return torch.squeeze(quantized_embeddings)

    def forward(self, x):
        return self.postprocess(x)


def make_layers():
    layers = []
    in_channels = 1
    for v in [64, "M", 128, "M", 256, 256, "M", 512, 512, "M"]:
        if v == "M":
            layers += [nn.MaxPool2d(kernel_size=2, stride=2)]
        else:
            conv2d = nn.Conv2d(in_channels, v, kernel_size=3, padding=1)
            layers += [conv2d, nn.ReLU(inplace=True)]
            in_channels = v
    return nn.Sequential(*layers)


class VGGish(VGG):
    def __init__(
        self, ckpt, pretrained=True, postprocess=True, progress=True, **kwargs
    ):
        super().__init__(make_layers())
        if pretrained:
            super().load_state_dict(ckpt["vggish"])

        self.postprocess = postprocess
        if self.postprocess:
            self.pproc = Postprocessor()
            if pretrained:
                # TODO: Convert the state_dict to torch
                ckpt["pca"][vggish_params.PCA_EIGEN_VECTORS_NAME] = torch.as_tensor(
                    ckpt["pca"][vggish_params.PCA_EIGEN_VECTORS_NAME], dtype=torch.float
                )
                ckpt["pca"][vggish_params.PCA_MEANS_NAME] = torch.as_tensor(
                    ckpt["pca"][vggish_params.PCA_MEANS_NAME].reshape(-1, 1),
                    dtype=torch.float,
                )
                self.pproc.load_state_dict(ckpt["pca"])

    def forward(self, x):
        x = super().forward(x)
        if self.postprocess:
            x = self._postprocess(x)
        return x

    def _postprocess(self, x):
        return self.pproc(x)



import logging
import os
import random
from dataclasses import dataclass
from typing import Callable, List

import torch
import torchaudio

from s3prl.dataio.encoder.category import CategoryEncoder
from s3prl.dataio.encoder.g2p import G2P
from s3prl.dataio.encoder.tokenizer import (
    Tokenizer,
    default_phoneme_tokenizer,
    load_tokenizer,
)
from s3prl.dataio.encoder.vocabulary import generate_vocab

from .base import AugmentedDynamicItemDataset, DataPipe

logger = logging.getLogger(__name__)


class SetOutputKeys(DataPipe):
    def __init__(self, output_keys: dict = None) -> None:
        super().__init__()
        self.output_keys = output_keys

    def forward(self, dataset: AugmentedDynamicItemDataset):
        dataset.update_output_keys(self.output_keys)
        return dataset


@dataclass
class LoadAudio(DataPipe):
    audio_sample_rate: int = 16000
    audio_channel_reduction: str = "first"
    sox_effects: list = None

    wav_path_name: str = "wav_path"
    wav_name: str = "wav"
    start_sec_name: str = "start_sec"
    end_sec_name: str = "end_sec"

    def load_audio(
        self,
        wav_path,
        start_sec: float = None,
        end_sec: float = None,
    ):
        crop_segment = start_sec is not None and end_sec is not None

        torchaudio.set_audio_backend("sox_io")
        wav, sr = torchaudio.load(
            wav_path,
            frame_offset=round(start_sec * self.audio_sample_rate)
            if crop_segment
            else 0,
            num_frames=round((end_sec - start_sec) * self.audio_sample_rate)
            if crop_segment
            else -1,
        )

        if self.sox_effects is not None:
            wav, sr = torchaudio.sox_effects.apply_effects_tensor(
                wav, sr, effects=self.sox_effects
            )

        if sr != self.audio_sample_rate:
            resampler = torchaudio.transforms.Resample(sr, self.audio_sample_rate)
            wav = resampler(wav)

        if self.audio_channel_reduction == "first":
            wav = wav[0]
        elif self.audio_channel_reduction == "mean":
            wav = wav.mean(dim=0)

        wav = wav.view(-1, 1)
        return wav

    def compute_length(self, wav):
        return len(wav)

    def forward(self, dataset: AugmentedDynamicItemDataset):
        item = dataset[0]
        if self.start_sec_name in item and self.end_sec_name in item:
            crop_segment = True
        else:
            crop_segment = False

        if not crop_segment:
            dataset.add_dynamic_item(
                self.load_audio, takes=self.wav_path_name, provides=self.wav_name
            )
        else:
            dataset.add_dynamic_item(
                self.load_audio,
                takes=[self.wav_path_name, self.start_sec_name, self.end_sec_name],
                provides=self.wav_name,
            )
        dataset.add_dynamic_item(
            self.compute_length,
            takes=self.wav_name,
            provides=f"{self.wav_name}_len",
        )
        return dataset


@dataclass
class EncodeCategory(DataPipe):
    train_category_encoder: bool = False
    label_name: str = "label"
    category_encoder_name: str = "category"
    encoded_target_name: str = "class_id"

    def prepare_category(self, labels):
        return CategoryEncoder(sorted(list(set(labels))))

    def encode_label(self, category, label):
        return category.encode(label)

    def forward(self, dataset: AugmentedDynamicItemDataset):
        if self.train_category_encoder:
            with dataset.output_keys_as([self.label_name]):
                labels = [item[self.label_name] for item in dataset]
            category = self.prepare_category(labels)
            dataset.add_tool(self.category_encoder_name, category)

        category = dataset.get_tool(self.category_encoder_name)
        dataset.add_tool("output_size", len(category))

        dataset.add_dynamic_item(
            self.encode_label,
            takes=[self.category_encoder_name, self.label_name],
            provides=self.encoded_target_name,
        )
        return dataset


@dataclass
class EncodeMultipleCategory(EncodeCategory):
    train_category_encoder: bool = False
    label_name: str = "labels"
    category_encoder_name: str = "categories"
    encoded_target_name: str = "class_ids"

    def encode_label(self, categories, labels):
        return torch.LongTensor(
            [category.encode(label) for category, label in zip(categories, labels)]
        )

    def forward(self, dataset: AugmentedDynamicItemDataset):
        if self.train_category_encoder:
            with dataset.output_keys_as([self.label_name]):
                labels = [item[self.label_name] for item in dataset]
            label_types = list(zip(*labels))
            categories = [
                self.prepare_category(label_type) for label_type in label_types
            ]
            dataset.add_tool(self.category_encoder_name, categories)
            dataset.add_tool("output_size", sum([len(c) for c in categories]))

        dataset.add_dynamic_item(
            self.encode_label,
            takes=[self.category_encoder_name, self.label_name],
            provides=self.encoded_target_name,
        )
        return dataset


@dataclass
class EncodeMultiLabel(DataPipe):
    label_name: str = "labels"
    category_encoder_name: str = "category"
    encoded_target_name: str = "binary_labels"

    @staticmethod
    def label_to_binary_vector(label: List, num_labels: int) -> torch.Tensor:
        # Lame special case for multilabel with no labels
        if len(label) == 0:
            # BCEWithLogitsLoss wants float not long targets
            binary_labels = torch.zeros((num_labels,), dtype=torch.float)
        else:
            binary_labels = torch.zeros((num_labels,)).scatter(
                0, torch.tensor(label), 1.0
            )

        # Validate the binary vector we just created
        assert set(torch.where(binary_labels == 1.0)[0].numpy()) == set(label)
        return binary_labels

    def encode_label(self, category, labels):
        labels = [category.encode(label) for label in labels]
        binary_labels = self.label_to_binary_vector(labels, len(category))
        return binary_labels

    def forward(self, dataset: AugmentedDynamicItemDataset):
        if not dataset.has_tool(self.category_encoder_name):
            with dataset.output_keys_as([self.label_name]):
                all_labels = []
                for item in dataset:
                    all_labels.extend(item[self.label_name])
                all_labels.sort()
                all_labels = set(all_labels)
                category = CategoryEncoder(all_labels)
                dataset.add_tool(self.category_encoder_name, category)

        category = dataset.get_tool(self.category_encoder_name)
        dataset.add_tool("output_size", len(category))

        dataset.add_dynamic_item(
            self.encode_label,
            takes=[self.category_encoder_name, self.label_name],
            provides=self.encoded_target_name,
        )
        return dataset


@dataclass
class GenerateTokenizer(DataPipe):
    generate: bool = True
    tokenizer_name: str = "tokenizer"
    text_name: str = "transcription"
    vocab_type: str = "character"
    text_file: str = None
    vocab_file: str = None
    slots_file: str = None
    vocab_args: dict = None

    def prepare_tokenizer(self, text_list: str = None) -> Tokenizer:
        """Generates tokenizer from text data.

        Args:
            text_list (str, optional): List of text. Defaults to None.

        Returns:
            Tokenizer: Generated tokenizer
        """

        vocab_args = self.vocab_args or {}
        assert isinstance(vocab_args, dict)

        if text_list is not None:
            vocab_result = generate_vocab(
                self.vocab_type, text_list=text_list, **vocab_args
            )
        else:
            vocab_result = generate_vocab(
                self.vocab_type, text_file=self.text_file, **vocab_args
            )
        vocab_list = vocab_result if isinstance(vocab_result, list) else None
        vocab_file = vocab_result if isinstance(vocab_result, str) else None

        tokenizer = load_tokenizer(
            self.vocab_type,
            vocab_file=vocab_file,
            vocab_list=vocab_list,
            slots_file=self.slots_file,
        )
        return tokenizer

    def forward(self, dataset: AugmentedDynamicItemDataset):
        try:
            tokenizer = dataset.get_tool(self.tokenizer_name)
            logger.info(
                f"Tokenizer (name = {self.tokenizer_name}) exists in dataset, skip generation."
            )
        except KeyError:
            if self.generate:
                if self.vocab_file is not None and os.path.exists(self.vocab_file):
                    tokenizer = load_tokenizer(
                        self.vocab_type,
                        vocab_file=self.vocab_file,
                        slots_file=self.slots_file,
                    )
                else:
                    text_list = None
                    if self.text_file is None:
                        with dataset.output_keys_as([self.text_name]):
                            text_list = [item[self.text_name] for item in dataset]

                    tokenizer = self.prepare_tokenizer(text_list)

                dataset.add_tool(self.tokenizer_name, tokenizer)
            else:
                logger.warning(
                    "No tokenizer is found or generated. No-op for this DataPipe"
                )

        return dataset


@dataclass
class EncodeText(DataPipe):
    text_name: str = "transcription"
    output_text_name: str = "tokenized_text"
    tokenizer_name: str = "tokenizer"

    def encode_text(self, tokenizer: Tokenizer, text: str) -> torch.LongTensor:
        return torch.LongTensor(tokenizer.encode(text))

    def forward(self, dataset: AugmentedDynamicItemDataset):
        try:
            tokenizer = dataset.get_tool(self.tokenizer_name)
        except KeyError:
            raise KeyError(f"Tokenizer (name = {self.tokenizer_name}) not found!")

        dataset.add_dynamic_item(
            self.encode_text,
            takes=[self.tokenizer_name, self.text_name],
            provides=self.output_text_name,
        )
        dataset.add_tool("output_size", tokenizer.vocab_size)

        return dataset


@dataclass
class Phonemize(DataPipe):
    text_name: str = "transcription"
    phonemized_text_name: str = "phonemized_text"
    output_text_name: str = "tokenized_text"
    g2p_name: str = "g2p"
    tokenizer_name: str = "tokenizer"

    def grapheme2phoneme(self, g2p: G2P, text: str) -> str:
        return g2p.encode(text)

    def encode_text(self, tokenizer: Tokenizer, text: str) -> torch.LongTensor:
        return torch.LongTensor(tokenizer.encode(text))

    def forward(self, dataset: AugmentedDynamicItemDataset):
        if not dataset.has_tool(self.g2p_name):
            logger.warning(
                f"Cannot find {self.g2p_name} in dataset, use default G2P instead."
            )
            dataset.add_tool(self.g2p_name, G2P())

        if not dataset.has_tool(self.tokenizer_name):
            logger.warning(
                f"Cannot find {self.tokenizer_name} in dataset, use default tokenizer instead."
            )
            dataset.add_tool(self.tokenizer_name, default_phoneme_tokenizer())

        dataset.add_dynamic_item(
            self.grapheme2phoneme,
            takes=[self.g2p_name, self.text_name],
            provides=self.phonemized_text_name,
        )

        dataset.add_dynamic_item(
            self.encode_text,
            takes=[self.tokenizer_name, self.phonemized_text_name],
            provides=self.output_text_name,
        )

        tokenizer = dataset.get_tool(self.tokenizer_name)
        dataset.add_tool("output_size", tokenizer.vocab_size)

        return dataset


@dataclass
class RandomCrop(DataPipe):
    """
    Completely randomized for every batch even with the same datapoint id.
    Only suitable for training.
    """

    sample_rate: int = 16000
    max_secs: float = None

    wav_name: str = "wav"
    crop_name: str = "wav_crop"

    def crop_wav(self, wav):
        if self.max_secs is not None and wav.size(0) > self.max_secs * self.sample_rate:
            start = random.randint(0, wav.size(0) - self.max_secs * self.sample_rate)
            end = start + self.max_secs * self.sample_rate
            wav = wav[round(start) : round(end)]
        return wav, wav.size(0)

    def forward(
        self, dataset: AugmentedDynamicItemDataset
    ) -> AugmentedDynamicItemDataset:
        dataset.add_dynamic_item(
            self.crop_wav,
            takes=[self.wav_name],
            provides=[self.crop_name, f"{self.crop_name}_len"],
        )
        return dataset



from s3prl.util.download import _urls_to_filepaths

from .expert import UpstreamExpert as _UpstreamExpert


def ast(
    refresh: bool = False,
    window_secs: float = 10.24,
    stride_secs: float = 10.24,
    **kwds,
):
    kwds["ckpt"] = _urls_to_filepaths(
        "https://www.dropbox.com/s/ca0b1v2nlxzyeb4/audioset_10_10_0.4593.pth?dl=1",
        refresh=refresh,
    )
    return _UpstreamExpert(window_secs=window_secs, stride_secs=stride_secs, **kwds)



"""
Pre-defined python recipes with customizable methods
"""

from .asr.superb_asr import SuperbASR
from .asr.superb_pr import SuperbPR
from .asr.superb_sf import SuperbSF
from .asv.superb_asv import SuperbASV
from .common.example import CommonExample
from .common.hear_beijing_opera import HearBeijingOpera
from .common.hear_cremad import HearCremaD
from .common.hear_dcase_2016_task2 import HearDcase2016Task2
from .common.hear_esc50 import HearESC50
from .common.hear_fsd import HearFSD
from .common.hear_gsc5hr import HearGSC5hr
from .common.hear_gtzan import HearGtzan
from .common.hear_gtzan_music_speech import HearGtzanMusicSpeech
from .common.hear_gunshot import HearGunshot
from .common.hear_libricount import HearLibriCount
from .common.hear_maestro import HearMaestro
from .common.hear_nsynth5hr import HearNsynth5hr
from .common.hear_stroke import HearStroke
from .common.hear_tonic import HearTonic
from .common.hear_vocal import HearVocal
from .common.hear_vox_lingual import HearVoxLingual
from .common.superb_er import SuperbER
from .common.superb_ic import SuperbIC
from .common.superb_ks import SuperbKS
from .common.superb_sid import SuperbSID
from .diarization.superb_sd import SuperbSD

__all__ = [
    "SuperbASR",
    "SuperbPR",
    "SuperbSF",
    "SuperbASV",
    "SuperbER",
    "SuperbIC",
    "SuperbKS",
    "SuperbSID",
    "SuperbSD",
    "HearFSD",
    "HearESC50",
    "HearBeijingOpera",
    "HearCremaD",
    "HearGSC5hr",
    "HearGtzanMusicSpeech",
    "HearGtzan",
    "HearGunshot",
    "HearLibriCount",
    "HearNsynth5hr",
    "HearStroke",
    "HearTonic",
    "HearVocal",
    "HearVoxLingual",
    "HearDcase2016Task2",
    "HearMaestro",
    "CommonExample",
]



"""
Group the data points with the same key into the same batch

Authors:
  * Leo 2022
"""

from collections import defaultdict
from typing import List

__all__ = [
    "GroupSameItemSampler",
]


class GroupSameItemSampler:
    def __init__(
        self,
        items: List[str],
    ) -> None:
        self.indices = defaultdict(list)
        for idx, item in enumerate(items):
            self.indices[item].append(idx)

        self.epoch = 0

    def set_epoch(self, epoch: int):
        self.epoch = epoch

    def __iter__(self):
        for batch_indices in self.indices.values():
            yield batch_indices

    def __len__(self):
        return len(list(iter(self)))



# --------------------------------------------------------
# LightHuBERT: Lightweight and Configurable Speech Representation Learning with Once-for-All Hidden-Unit BERT (https://arxiv.org/pdf/2203.15610.pdf)
# Github source: https://github.com/mechanicalsea/lighthubert
# Licensed under The MIT License [see LICENSE for details]
# Based on fairseq code bases
# https://github.com/pytorch/fairseq
# --------------------------------------------------------

import logging
import random
from typing import Dict, List, Optional, Tuple

import numpy as np
import torch
import torch.nn as nn

from .modules.fairseq_modules import GradMultiply
from .modules.scaling_linear import SLinear
from .modules.scaling_multihead import SMHA
from .modules.scaling_transformer import STransformerEncoder
from .modules.w2v2_modules import ConvFeatureExtractionModel

logger = logging.getLogger(__name__)


def compute_mask_indices(
    shape: Tuple[int, int],
    padding_mask: Optional[torch.Tensor],
    mask_prob: float,
    mask_length: int,
    mask_type: str = "static",
    mask_other: float = 0.0,
    min_masks: int = 0,
    no_overlap: bool = False,
    min_space: int = 0,
    require_same_masks: bool = True,
    mask_dropout: float = 0.0,
) -> np.ndarray:
    """
    Computes random mask spans for a given shape

    Args:
        shape: the the shape for which to compute masks.
            should be of size 2 where first element is batch size and 2nd is timesteps
        padding_mask: optional padding mask of the same size as shape, which will prevent masking padded elements
        mask_prob: probability for each token to be chosen as start of the span to be masked. this will be multiplied by
            number of timesteps divided by length of mask span to mask approximately this percentage of all elements.
            however due to overlaps, the actual number will be smaller (unless no_overlap is True)
        mask_type: how to compute mask lengths
            static = fixed size
            uniform = sample from uniform distribution [mask_other, mask_length*2]
            normal = sample from normal distribution with mean mask_length and stdev mask_other. mask is min 1 element
            poisson = sample from possion distribution with lambda = mask length
        min_masks: minimum number of masked spans
        no_overlap: if false, will switch to an alternative recursive algorithm that prevents spans from overlapping
        min_space: only used if no_overlap is True, this is how many elements to keep unmasked between spans
        require_same_masks: if true, will randomly drop out masks until same amount of masks remains in each sample
        mask_dropout: randomly dropout this percentage of masks in each example
    """

    bsz, all_sz = shape
    mask = np.full((bsz, all_sz), False)

    all_num_mask = int(
        # add a random number for probabilistic rounding
        mask_prob * all_sz / float(mask_length)
        + np.random.rand()
    )

    all_num_mask = max(min_masks, all_num_mask)

    mask_idcs = []
    for i in range(bsz):
        if padding_mask is not None:
            sz = all_sz - padding_mask[i].long().sum().item()
            num_mask = int(
                # add a random number for probabilistic rounding
                mask_prob * sz / float(mask_length)
                + np.random.rand()
            )
            num_mask = max(min_masks, num_mask)
        else:
            sz = all_sz
            num_mask = all_num_mask

        if mask_type == "static":
            lengths = np.full(num_mask, mask_length)
        elif mask_type == "uniform":
            lengths = np.random.randint(mask_other, mask_length * 2 + 1, size=num_mask)
        elif mask_type == "normal":
            lengths = np.random.normal(mask_length, mask_other, size=num_mask)
            lengths = [max(1, int(round(x))) for x in lengths]
        elif mask_type == "poisson":
            lengths = np.random.poisson(mask_length, size=num_mask)
            lengths = [int(round(x)) for x in lengths]
        else:
            raise Exception("unknown mask selection " + mask_type)

        if sum(lengths) == 0:
            lengths[0] = min(mask_length, sz - 1)

        if no_overlap:
            mask_idc = []

            def arrange(s, e, length, keep_length):
                span_start = np.random.randint(s, e - length)
                mask_idc.extend(span_start + i for i in range(length))

                new_parts = []
                if span_start - s - min_space >= keep_length:
                    new_parts.append((s, span_start - min_space + 1))
                if e - span_start - keep_length - min_space > keep_length:
                    new_parts.append((span_start + length + min_space, e))
                return new_parts

            parts = [(0, sz)]
            min_length = min(lengths)
            for length in sorted(lengths, reverse=True):
                lens = np.fromiter(
                    (e - s if e - s >= length + min_space else 0 for s, e in parts),
                    np.int,
                )
                l_sum = np.sum(lens)
                if l_sum == 0:
                    break
                probs = lens / np.sum(lens)
                c = np.random.choice(len(parts), p=probs)
                s, e = parts.pop(c)
                parts.extend(arrange(s, e, length, min_length))
            mask_idc = np.asarray(mask_idc)
        else:
            min_len = min(lengths)
            if sz - min_len <= num_mask:
                min_len = sz - num_mask - 1

            mask_idc = np.random.choice(sz - min_len, num_mask, replace=False)

            mask_idc = np.asarray(
                [
                    mask_idc[j] + offset
                    for j in range(len(mask_idc))
                    for offset in range(lengths[j])
                ]
            )

        mask_idcs.append(np.unique(mask_idc[mask_idc < sz]))

    min_len = min([len(m) for m in mask_idcs])
    for i, mask_idc in enumerate(mask_idcs):
        if len(mask_idc) > min_len and require_same_masks:
            mask_idc = np.random.choice(mask_idc, min_len, replace=False)
        if mask_dropout > 0:
            num_holes = np.rint(len(mask_idc) * mask_dropout).astype(int)
            mask_idc = np.random.choice(
                mask_idc, len(mask_idc) - num_holes, replace=False
            )

        mask[i, mask_idc] = True

    return mask


class LightHuBERTSupernetConfig(object):
    """LightHuBERT search space providing supernet, search space, a subnet."""

    def __init__(self, supernet_type="base"):
        assert supernet_type.lower() in ["base", "small"]
        self.supernet_type = supernet_type.lower()

        def prod(x: List):
            ans = 1
            for xi in x:
                ans *= len(set(xi))
            return ans

        self.search_space_size = self.search_space["embed_dim"].__len__() * (
            sum(
                len(self.search_space["ffn_ratio"]) ** di
                * len(self.search_space["heads_num"]) ** di
                * prod(
                    self.search_space["slide_wsz"][:di]
                    if "slide_wsz" in self.search_space
                    else [["global"] for _ in range(di)]
                )
                for di in self.search_space["layer_num"]
            )
        )

    @property
    def supernet(self):
        return {
            "atten_dim": 768,
            "embed_dim": 768,
            "ffn_ratio": 4.0,
            "heads_num": 12,
            "layer_num": 12,
        }

    @property
    def search_space(self):
        if self.supernet_type == "base":
            return {
                "atten_dim": [512, 640, 768],
                "embed_dim": [512, 640, 768],
                "ffn_ratio": [3.5, 4.0],
                "heads_num": [8, 10, 12],
                "layer_num": [12],
            }
        elif self.supernet_type == "small":
            return {
                "atten_dim": [256, 384, 512],
                "embed_dim": [256, 384, 512],
                "ffn_ratio": [3.0, 3.5, 4.0],
                "heads_num": [4, 6, 8],
                "layer_num": [10, 11, 12],
            }

    @property
    def subnet(self):
        if self.supernet_type == "base":
            return {
                "atten_dim": [640 for _ in range(12)],
                "embed_dim": 640,
                "ffn_ratio": [4.0 for _ in range(12)],
                "ffn_embed": [2560 for _ in range(12)],
                "heads_num": [10 for _ in range(12)],
                "layer_num": 12,
                "slide_wsz": ["global" for _ in range(12)],
            }
        elif self.supernet_type == "small":
            return {
                "atten_dim": [384 for _ in range(12)],
                "embed_dim": 384,
                "ffn_ratio": [4.0 for _ in range(12)],
                "ffn_embed": [1536 for _ in range(12)],
                "heads_num": [6 for _ in range(12)],
                "layer_num": 12,
                "slide_wsz": ["global" for _ in range(12)],
            }

    @property
    def dimensions(self) -> list:
        return ["atten_dim", "embed_dim", "ffn_ratio", "heads_num", "layer_num"]

    @property
    def num_subnets(self) -> int:
        """Count the number of subnets in the supernet."""
        return self.search_space_size

    @property
    def slide_mode(self) -> str:
        return "stride"

    @property
    def max_subnet(self) -> dict:
        layer_num = max(self.search_space["layer_num"])
        embed_dim = max(self.search_space["embed_dim"])
        ffn_embed = [
            int(max(self.search_space["ffn_ratio"]) * embed_dim)
            for _ in range(layer_num)
        ]
        heads_num = [max(self.search_space["heads_num"]) for _ in range(layer_num)]
        atten_dim = [heads_num[i] * 64 for i in range(layer_num)]
        slide_wsz = ["global" for _ in range(layer_num)]
        return {
            "atten_dim": atten_dim,  # List[int]
            "embed_dim": embed_dim,  # int
            "ffn_embed": ffn_embed,  # List[int]
            "heads_num": heads_num,  # List[int]
            "layer_num": layer_num,  # int
            "slide_wsz": slide_wsz,  # List[int] or List[str]
        }

    @property
    def min_subnet(self) -> dict:
        assert self.search_space is not None
        assert self.search_space is not None
        layer_num = min(self.search_space["layer_num"])
        embed_dim = min(self.search_space["embed_dim"])
        ffn_embed = [
            int(min(self.search_space["ffn_ratio"]) * embed_dim)
            for _ in range(layer_num)
        ]
        heads_num = [min(self.search_space["heads_num"]) for _ in range(layer_num)]
        atten_dim = [heads_num[i] * 64 for i in range(layer_num)]
        slide_wsz = ["global" for _ in range(layer_num)]
        return {
            "atten_dim": atten_dim,  # List[int]
            "embed_dim": embed_dim,  # int
            "ffn_embed": ffn_embed,  # List[int]
            "heads_num": heads_num,  # List[int]
            "layer_num": layer_num,  # int
            "slide_wsz": slide_wsz,  # List[int] or List[str]
        }

    def sample_subnet(self) -> dict:
        """sample a subnet for search space, e.g.,
        {
            "atten_dim": [512, 512, 512, 512, 512, 512]        # List[int]
            "embed_dim": 512,                                  # int
            # "ffn_ratio": [3.0, 3.0, 3.0, 3.0, 3.0, 3.0]      # List[int]
            "ffn_embed": [1536, 1536, 1536, 1536, 1536, 1536]  # List[int]
            "heads_num": [8, 8, 8, 8, 8, 8, 8],                # List[int]
            "layer_num": 6,                                    # int
            "slide_wsz": [global, global, global, global, global, global], # List[int] or List[str]
        }
        ffn_ratio 3.0 -> ffn_embed 1536
        heads_num 8 -> atten_dim 512
        """
        assert self.search_space is not None
        layer_num = random.choice(self.search_space["layer_num"])
        embed_dim = random.choice(self.search_space["embed_dim"])
        ffn_embed = [
            int(random.choice(self.search_space["ffn_ratio"]) * embed_dim)
            for _ in range(layer_num)
        ]
        heads_num = [
            random.choice(self.search_space["heads_num"]) for _ in range(layer_num)
        ]
        atten_dim = [heads_num[i] * 64 for i in range(layer_num)]
        if "slide_wsz" in self.search_space:
            slide_wsz = [
                random.choice(self.search_space["slide_wsz"][i])
                for i in range(layer_num)
            ]
            slide_wsz = [
                int(swsz) if swsz != "global" else str(swsz) for swsz in slide_wsz
            ]
        else:
            slide_wsz = ["global" for _ in range(layer_num)]
        return {
            "atten_dim": atten_dim,  # List[int]
            "embed_dim": embed_dim,  # int
            "ffn_embed": ffn_embed,  # List[int]
            "heads_num": heads_num,  # List[int]
            "layer_num": layer_num,  # int
            "slide_wsz": slide_wsz,  # List[int] or List[str]
        }


class LightHuBERTConfig:
    def __init__(self, cfg=None):
        self.supernet_type: str = "base"  # 'base' or 'small' supernet

        self.prune_encoder_pos_conv: bool = True  # if true, prune multi-layer position convolution totally, otherwise, prune last layer to encoder

        # Configuration for Student
        self.teacher_embed_dim: int = 768  # embedding size of teacher's outputs
        self.layer_pred_num: str = "0,0,0,0,0,0,0,0,0,0,0,1"  # specify the predicted number of each layer, e.g., '2,2,2,2,2,2'

        # add pos_conv_depth to hubert as wav2vec2 and enable data2vec architecture
        self.pos_conv_depth: int = 1  # depth of positional encoder network

        # hubert config
        self.label_rate: int = 50
        self.extractor_mode: str = "layer_norm"  # mode for feature extractor. default has a single group norm with d groups in the first conv block, whereas layer_norm has layer norms in every block (meant to use with normalize=True)
        self.encoder_layers: int = 12  # num encoder layers in the transformer
        self.encoder_embed_dim: int = 768  # encoder embedding dimension
        self.encoder_ffn_embed_dim: int = 3072  # encoder embedding dimension for FFN
        self.encoder_attention_heads: int = 12  # num encoder attention heads
        self.activation_fn: str = "gelu"  # activation function to use
        self.layer_type: str = "transformer"  # layer type in encoder

        # dropouts
        self.dropout: float = 0.1  # dropout probability for the transformer
        self.attention_dropout: float = 0.1  # dropout probability for attention weights
        self.activation_dropout: float = (
            0.0  # dropout probability after activation in FFN
        )
        self.encoder_layerdrop: float = (
            0.0  # probability of dropping a tarnsformer layer
        )
        self.dropout_input: float = (
            0.0  # dropout to apply to the input (after feat extr)
        )
        self.dropout_features: float = (
            0.0  # dropout to apply to the features (after feat extr)
        )

        self.layer_norm_first: bool = False  # apply layernorm first in the transformer
        self.conv_feature_layers: str = "[(512,10,5)] + [(512,3,2)] * 4 + [(512,2,2)] * 2"  # string describing convolutional feature extraction layers in form of a python list that contains [(dim, kernel_size, stride), ...]
        self.conv_bias: bool = False  # include bias in conv encoder
        self.feature_grad_mult: float = (
            1.0  # multiply feature extractor var grads by this
        )

        # masking
        self.mask_length: int = 10  # mask length"
        self.mask_prob: float = 0.65  # probability of replacing a token with mask
        self.mask_selection: str = "static"  # how to choose mask length
        self.mask_other: float = 0  # secondary mask argument (used for more complex distributions), see help in compute_mask_indicesh
        self.no_mask_overlap: bool = False  # whether to allow masks to overlap
        self.mask_min_space: int = (
            1  # min space between spans (if no overlap is enabled)
        )

        # channel masking
        self.mask_channel_length: int = 10  # length of the mask for features (channels)
        self.mask_channel_prob: float = 0.0  # probability of replacing a feature with 0
        self.mask_channel_selection: str = (
            "static"  # how to choose mask length for channel masking
        )
        self.mask_channel_other: float = 0  # secondary mask argument (used for more complex distributions), see help in compute_mask_indicesh
        self.no_mask_channel_overlap: bool = (
            False  # whether to allow channel masks to overlap
        )
        self.mask_channel_min_space: int = (
            1  # min space between spans (if no overlap is enabled)
        )

        # positional embeddings
        self.conv_pos: int = (
            128  # number of filters for convolutional positional embeddings
        )
        self.conv_pos_groups: int = (
            16  # number of groups for convolutional positional embedding
        )

        # FP16 optimization
        self.required_seq_len_multiple: int = 2  # pad the input to encoder such that the sequence length is divisible by multiple

        if cfg is not None:
            self.update(cfg)

    def update(self, cfg: dict):
        for ki, vi in cfg.items():
            if ki in self.__dict__:
                self.__dict__.update({ki: vi})


class LightHuBERT(nn.Module):
    def __init__(self, cfg: LightHuBERTConfig):
        super().__init__()
        logger.info(f"LightHuBERT Config: {cfg}")
        self.cfg = cfg

        feature_enc_layers = eval(cfg.conv_feature_layers)  # noqa
        self.embed = feature_enc_layers[-1][0]

        self.feature_extractor = ConvFeatureExtractionModel(
            conv_layers=feature_enc_layers,
            dropout=0.0,
            mode=cfg.extractor_mode,
            conv_bias=cfg.conv_bias,
        )

        self.post_extract_proj = (
            SLinear(self.embed, cfg.encoder_embed_dim)
            if self.embed != cfg.encoder_embed_dim
            else None
        )  # lighthubert component

        self.mask_prob = cfg.mask_prob
        self.mask_selection = cfg.mask_selection
        self.mask_other = cfg.mask_other
        self.mask_length = cfg.mask_length
        self.no_mask_overlap = cfg.no_mask_overlap
        self.mask_min_space = cfg.mask_min_space

        self.mask_channel_prob = cfg.mask_channel_prob
        self.mask_channel_selection = cfg.mask_channel_selection
        self.mask_channel_other = cfg.mask_channel_other
        self.mask_channel_length = cfg.mask_channel_length
        self.no_mask_channel_overlap = cfg.no_mask_channel_overlap
        self.mask_channel_min_space = cfg.mask_channel_min_space

        self.dropout_input = nn.Dropout(cfg.dropout_input)
        self.dropout_features = nn.Dropout(cfg.dropout_features)

        self.feature_grad_mult = cfg.feature_grad_mult

        self.mask_emb = nn.Parameter(
            torch.FloatTensor(cfg.encoder_embed_dim).uniform_()
        )

        self.encoder = STransformerEncoder(cfg)  # lighthubert component
        self.layer_norm = nn.LayerNorm(self.embed)

        # student config
        self.encoder_layers = cfg.encoder_layers
        self.encoder_embed_dim = cfg.encoder_embed_dim
        self.teacher_embed_dim = cfg.teacher_embed_dim

        layer_pred_num = [max(int(n), 0) for n in cfg.layer_pred_num.split(",")]
        assert len(layer_pred_num) == self.encoder_layers, f"{len(layer_pred_num)}"
        self.layer_pred_heads = torch.nn.ModuleDict()
        for layer_i in range(self.encoder_layers):
            if layer_pred_num[layer_i] > 0:
                self.layer_pred_heads[f"{layer_i}"] = SLinear(
                    self.encoder_embed_dim,
                    self.teacher_embed_dim * layer_pred_num[layer_i],
                    bias=False,
                )
        self.layer_pred_num = layer_pred_num
        logger.info(f"predicting heads: {layer_pred_num}")

        # lighthubert config
        assert cfg.encoder_attention_heads * 64 == cfg.encoder_embed_dim
        self.sample_layer_num = cfg.encoder_layers
        self.sample_embed_dim = cfg.encoder_embed_dim
        self.sample_ffn_embed = cfg.encoder_ffn_embed_dim
        self.sample_heads_num = cfg.encoder_attention_heads
        self.sample_atten_dim = cfg.encoder_embed_dim
        self.sample_slide_wsz = None

        def subnet_params(subnet):
            self.set_sample_config(subnet)
            return self.calc_sampled_param_num()

        self.supernet = LightHuBERTSupernetConfig(cfg.supernet_type)
        self._switch_slide_attention()
        self.dynamize(False)

        # search space
        logger.info(
            f"search space ({self.supernet.num_subnets:,} subnets): {self.supernet.search_space}"
        )
        # min subnet
        total_params = subnet_params(self.supernet.min_subnet)
        logger.info(
            f"min subnet ({total_params/1e6:.0f} Params): {self.supernet.min_subnet}"
        )
        # max subnet
        total_params = subnet_params(self.supernet.max_subnet)
        logger.info(
            f"max subnet ({total_params/1e6:.0f} Params): {self.supernet.max_subnet}"
        )

    def _switch_slide_attention(self, slide_mode="stride"):
        """Set sliding attention manner to either stride or mask."""
        assert slide_mode in ["stride", "mask"]
        for n, m in self.named_modules():
            if isinstance(m, SMHA):
                assert hasattr(m, "set_slide_mode")
                m.set_slide_mode(slide_mode)

    def dynamize(self, mode: bool = True, log_subnet: bool = True):
        """To determine whether to sample a subnet during forward.
        If not training, self.dynamize(False)
        Else: self.dynamize()

        self.feature_extractor: ConvFeatureExtractionModel
        self.post_extract_proj: nn.Linear
        self.mask_emb: nn.Parameter
        self.encoder: TransformerEncoder
        self.layer_norm: nn.LayerNorm
        """
        self.dynamic = mode
        self.verbose = log_subnet & (self.supernet.num_subnets > 1)
        if not self.dynamic:
            subnet = self.supernet.subnet
            self.set_sample_config(subnet)
            if hasattr(self, "handle"):
                self.handle.remove()
            return

        # sample a subnet before forward
        def set_subnet(module, input):
            if not getattr(module, "dynamic", False):
                return
            assert hasattr(module, "set_sample_config")
            subnet = module.supernet.sample_subnet()
            module.set_sample_config(subnet)
            if getattr(module, "verbose", False):
                total_params = module.calc_sampled_param_num() / 1e6
                logger.info(f"dynamic subnet ({total_params:.2f}M Params): {subnet}")

        self.handle = self.register_forward_pre_hook(set_subnet)

    def set_sample_config(self, config: dict):
        """
        config: {
            "atten_dim": atten_dim, # List[int]
            "embed_dim": embed_dim, # int
            "ffn_embed": ffn_embed, # List[int]
            "heads_num": heads_num, # List[int]
            "layer_num": layer_num, # int
            "slide_wsz": slide_wsz, # List[int] or List[str]
        }
        """
        self.sample_layer_num = config["layer_num"]
        self.sample_atten_dim = config["atten_dim"]
        self.sample_embed_dim = config["embed_dim"]
        self.sample_ffn_embed = config["ffn_embed"]
        self.sample_heads_num = config["heads_num"]
        self.sample_slide_wsz = config["slide_wsz"]
        self._sample_parameters()

    def _sample_parameters(self):
        self.post_extract_proj.set_sample_config(
            self.embed,
            self.sample_embed_dim,
        )
        self.encoder.set_sample_config(
            self.sample_layer_num,
            self.sample_atten_dim,
            self.sample_embed_dim,
            self.sample_ffn_embed,
            self.sample_heads_num,
            self.sample_slide_wsz,
        )
        if self.layer_pred_heads is not None:
            for _, lrhi in self.layer_pred_heads.items():
                lrhi.set_sample_config(self.sample_embed_dim, self.teacher_embed_dim)

    def calc_sampled_param_num(self):
        total_params = 0
        for n, p in self.named_parameters():
            if (
                not n.startswith("post_extract_proj")
                and not n.startswith("encoder")
                and not n.startswith("layer_pred_heads")
            ):
                total_params += p.numel()
        total_params += self.post_extract_proj.calc_sampled_param_num()
        total_params += self.encoder.calc_sampled_param_num()
        # not consider prediction head
        # if self.layer_pred_heads is not None:
        #     for _, lrhi in self.layer_pred_heads.items():
        #         total_params += lrhi.calc_sampled_param_num()
        return total_params

    def apply_mask(self, x, padding_mask, target_list, mask_indices=None):
        """Refactor mask method to enable to mask with the given masking indices."""
        B, T, C = x.shape
        if self.mask_prob > 0:
            if mask_indices is None:
                mask_indices = compute_mask_indices(
                    (B, T),
                    padding_mask,
                    self.mask_prob,
                    self.mask_length,
                    self.mask_selection,
                    self.mask_other,
                    min_masks=2,
                    no_overlap=self.no_mask_overlap,
                    min_space=self.mask_min_space,
                )
                mask_indices = torch.from_numpy(mask_indices).to(x.device)
            x[mask_indices] = self.mask_emb[:C]
        else:
            mask_indices = None

        if self.mask_channel_prob > 0:
            mask_channel_indices = compute_mask_indices(
                (B, C),
                None,
                self.mask_channel_prob,
                self.mask_channel_length,
                self.mask_channel_selection,
                self.mask_channel_other,
                no_overlap=self.no_mask_channel_overlap,
                min_space=self.mask_channel_min_space,
            )
            mask_channel_indices = (
                torch.from_numpy(mask_channel_indices)
                .to(x.device)
                .unsqueeze(1)
                .expand(-1, T, -1)
            )
            x[mask_channel_indices] = 0

        return x, mask_indices

    def forward_features(self, source: torch.Tensor) -> torch.Tensor:
        if self.feature_grad_mult > 0:
            features = self.feature_extractor(source)
            if self.feature_grad_mult != 1.0:
                features = GradMultiply.apply(features, self.feature_grad_mult)
        else:
            with torch.no_grad():
                features = self.feature_extractor(source)
        return features

    def forward_padding_mask(
        self,
        features: torch.Tensor,
        padding_mask: torch.Tensor,
    ) -> torch.Tensor:
        extra = padding_mask.size(1) % features.size(1)
        if extra > 0:
            padding_mask = padding_mask[:, :-extra]
        padding_mask = padding_mask.view(padding_mask.size(0), features.size(1), -1)
        padding_mask = padding_mask.all(-1)
        return padding_mask

    def forward(
        self,
        source: torch.Tensor,
        target_list: Optional[List[torch.Tensor]] = None,
        padding_mask: Optional[torch.Tensor] = None,
        mask: bool = True,
        features_only: bool = False,
        output_layer: Optional[int] = None,
        mask_indices=None,
    ) -> Dict[str, torch.Tensor]:
        """Refactor extract feature method to enable to output
        1. hidden representations
        2. mask indices corresponding to masked language modeling
        """
        features = self.forward_features(source)
        if target_list is not None:
            features, target_list = self.forward_targets(features, target_list)

        features_pen = features.float().pow(2).mean()

        features = features.transpose(1, 2)
        features = self.layer_norm(features)
        unmasked_features = features.clone()

        if padding_mask is not None:
            padding_mask = self.forward_padding_mask(features, padding_mask)

        if self.post_extract_proj is not None:
            features = self.post_extract_proj(features)

        features = self.dropout_input(features)
        unmasked_features = self.dropout_features(unmasked_features)

        if mask:
            x, mask_indices = self.apply_mask(
                features, padding_mask, target_list, mask_indices=mask_indices
            )
        else:
            x = features
            mask_indices = None

        # feature: (B, T, D), float
        # target: (B, T), long
        # x: (B, T, D), float
        # padding_mask: (B, T), bool
        # mask_indices: (B, T), bool
        x, layer_results = self.encoder(
            x,
            padding_mask=padding_mask,
            layer=None,
        )

        layer_head_list = [
            None,
        ] * self.encoder_layers
        if self.layer_pred_heads is not None:
            if (
                len(self.layer_pred_heads) == 1
                and f"{self.encoder_layers - 1}" in self.layer_pred_heads
            ):
                # enable layerdrop in training while only consider the last hidden state
                _layer_results = [
                    (None, None, None) for _ in range(self.encoder_layers - 1)
                ] + [(x.transpose(0, 1), None, None)]
            else:
                if len(layer_results) == self.encoder_layers:
                    _layer_results = layer_results
                else:
                    raise ValueError("Multi-layer prediction requires 0.0 layerdrop.")
            for layer_i, (layer_hs, _, _) in enumerate(_layer_results):
                if f"{layer_i}" in self.layer_pred_heads:
                    layer_pred_head = self.layer_pred_heads[f"{layer_i}"]
                    # enable hidden states with variable embedding size
                    layer_pred_head.set_sample_config(
                        layer_hs.size(-1), layer_pred_head.out_features
                    )
                    layer_output = (
                        layer_pred_head(layer_hs).transpose(0, 1).contiguous()
                    )
                    layer_head_list[layer_i] = torch.split(
                        layer_output, self.teacher_embed_dim, dim=-1
                    )

        hidden_states = [features] + [
            hs.transpose(0, 1).contiguous() for hs, attn, ls in layer_results
        ]
        attn_matrices = [
            attn for hs, attn, ls in layer_results
        ]  # List[(heads_num, batch_size, tgt_len, src_len)]

        if features_only:
            return {
                "x": x,
                "padding_mask": padding_mask,
                "hidden_states": hidden_states,
                "layer_heads": layer_head_list,
            }

        result = {
            "padding_mask": padding_mask,
            "features_pen": features_pen,
            "x": x,
            "hidden_states": hidden_states,
            "attn_matrices": attn_matrices,
            "layer_heads": layer_head_list,
            "mask_indices": mask_indices,
        }
        return result

    def extract_features(
        self,
        source: torch.Tensor,
        padding_mask: Optional[torch.Tensor] = None,
        mask: bool = False,
        ret_conv: bool = False,
        ret_hs: bool = False,
        cat_heads: bool = False,
        output_layer: Optional[int] = None,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """Refactor extract feature method to enable to output hidden representations."""
        res = self.forward(
            source,
            padding_mask=padding_mask,
            mask=mask,
            features_only=True,
            output_layer=output_layer,
        )
        if ret_hs:
            feature = res["hidden_states"]
            if cat_heads:
                layer_heads_list = [
                    layer_heads
                    for layer_heads in res["layer_heads"]
                    if layer_heads is not None
                ]
                for layer_heads in layer_heads_list:
                    feature += layer_heads
        elif ret_conv:
            feature = res["hidden_states"][0]
        else:
            feature = res["x"]
        return feature, res["padding_mask"]

    def freeze_feature_extractor(self):
        for param in self.feature_extractor.parameters():
            param.requires_grad = False
        self.extractor_frozen = True

    def get_extra_losses(self, net_output):
        extra_losses = []
        names = []

        if "features_pen" in net_output:
            if self.feature_extractor.parameters().__next__().requires_grad:
                extra_losses.append(net_output["features_pen"])
                names.append("features_pen")

        return extra_losses, names

    def remove_pretraining_modules(self):
        """Meantime remove predicting heads"""
        self.layer_pred_heads = None
        self.layer_pred_num = [
            0,
        ] * self.encoder_layers



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/wavlm/hubconf.py ]
#   Synopsis     [ the WavLM torch hubconf ]
#   Author       [ Microsoft ]
"""*********************************************************************************************"""


import os

from s3prl.util.download import _urls_to_filepaths

from .expert import UpstreamExpert as _UpstreamExpert


def wavlm_local(ckpt, *args, **kwargs):
    """
    The model from local ckpt
        ckpt (str): PATH
    """
    assert os.path.isfile(ckpt)
    return _UpstreamExpert(ckpt, *args, **kwargs)


def wavlm_url(ckpt, refresh=False, *args, **kwargs):
    """
    The model from google drive id
        ckpt (str): URL
        refresh (bool): whether to download ckpt/config again if existed
    """
    return wavlm_local(_urls_to_filepaths(ckpt, refresh=refresh), *args, **kwargs)


def wavlm(refresh=False, *args, **kwargs):
    """
    The default model - Base-Plus
        refresh (bool): whether to download ckpt/config again if existed
    """
    return wavlm_base_plus(refresh=refresh, *args, **kwargs)


def wavlm_base(refresh=False, *args, **kwargs):
    """
    The Base model
        refresh (bool): whether to download ckpt/config again if existed
    """
    kwargs[
        "ckpt"
    ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/wavlm_base.pt"

    return wavlm_url(refresh=refresh, *args, **kwargs)


def wavlm_base_plus(refresh=False, *args, **kwargs):
    """
    The Base-Plus model
        refresh (bool): whether to download ckpt/config again if existed
    """
    kwargs[
        "ckpt"
    ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/wavlm_base_plus.pt"

    return wavlm_url(refresh=refresh, *args, **kwargs)


def wavlm_large(refresh=False, *args, **kwargs):
    """
    The Large model
        refresh (bool): whether to download ckpt/config again if existed
    """
    kwargs[
        "ckpt"
    ] = "https://huggingface.co/s3prl/converted_ckpts/resolve/main/wavlm_large.pt"

    return wavlm_url(refresh=refresh, *args, **kwargs)



from s3prl.nn.linear import FrameLevelLinear
from s3prl.nn.specaug import ModelWithSpecaug


def test_specaug_model():
    model = FrameLevelLinear(input_size=13, output_size=25, hidden_size=32)
    model = ModelWithSpecaug(model)
    assert model.specaug.apply_time_mask == True
    assert model.specaug.apply_freq_mask == True



import json
import logging
import pickle
from collections import defaultdict
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path

import pandas as pd
import torch
import torchaudio
from omegaconf import MISSING

from s3prl.dataio.dataset import FrameLabelDataset, get_info
from s3prl.dataio.sampler import FixedBatchSizeBatchSampler, GroupSameItemSampler
from s3prl.task.event_prediction import EventPredictionTask

from ._hear_util import resample_hear_corpus
from .hear_fsd import HearFSD

logger = logging.getLogger(__name__)

__all__ = [
    "HearDcase2016Task2",
]


def dcase_2016_task2(
    target_dir: str,
    cache_dir: str,
    dataset_root: str,
    get_path_only: bool = False,
):
    target_dir: Path = Path(target_dir)
    train_csv = target_dir / "train.csv"
    valid_csv = target_dir / "valid.csv"
    test_csv = target_dir / "test.csv"

    if get_path_only:
        return train_csv, valid_csv, [test_csv]

    resample_hear_corpus(dataset_root, target_sr=16000)

    dataset_root = Path(dataset_root)
    wav_root: Path = dataset_root / "16000"

    def json_to_csv(json_path: str, csv_path: str, split: str):
        with open(json_path) as fp:
            metadata = json.load(fp)

        data = defaultdict(list)
        for utt in metadata:
            wav_path: Path = (wav_root / split / utt).resolve()
            assert wav_path.is_file()
            info = torchaudio.info(wav_path)
            baseinfo = {
                "record_id": utt,
                "wav_path": str(wav_path),
                "duration": info.num_frames / info.sample_rate,
            }
            for segment in metadata[utt]:
                fullinfo = deepcopy(baseinfo)
                fullinfo[
                    "utt_id"
                ] = f"{baseinfo['record_id']}-{int(segment['start'])}-{int(segment['end'])}"
                fullinfo["labels"] = segment["label"]
                fullinfo["start_sec"] = segment["start"] / 1000
                fullinfo["end_sec"] = segment["end"] / 1000

                for key, value in fullinfo.items():
                    data[key].append(value)

        pd.DataFrame(data=data).to_csv(csv_path, index=False)

    json_to_csv(dataset_root / "train.json", train_csv, "train")
    json_to_csv(dataset_root / "valid.json", valid_csv, "valid")
    json_to_csv(dataset_root / "test.json", test_csv, "test")

    return train_csv, valid_csv, [test_csv]


class HearDcase2016Task2(HearFSD):
    def default_config(self) -> dict:
        return dict(
            start=0,
            stop=None,
            target_dir=MISSING,
            cache_dir=None,
            remove_all_cache=False,
            prepare_data=dict(
                dataset_root=MISSING,
            ),
            build_dataset=dict(
                train=dict(
                    chunk_secs=4.0,
                    step_secs=4.0,
                ),
                valid=dict(
                    chunk_secs=4.0,
                    step_secs=4.0,
                ),
                test=dict(
                    chunk_secs=4.0,
                    step_secs=4.0,
                ),
            ),
            build_batch_sampler=dict(
                train=dict(
                    batch_size=5,
                    shuffle=True,
                ),
            ),
            build_upstream=dict(
                name=MISSING,
            ),
            build_featurizer=dict(
                layer_selections=None,
                normalize=False,
            ),
            build_downstream=dict(
                hidden_layers=2,
            ),
            build_model=dict(
                upstream_trainable=False,
            ),
            build_task=dict(
                prediction_type="multilabel",
                scores=["event_onset_200ms_fms", "segment_1s_er"],
                postprocessing_grid={
                    "median_filter_ms": [250],
                    "min_duration": [125, 250],
                },
            ),
            build_optimizer=dict(
                name="Adam",
                conf=dict(
                    lr=1.0e-3,
                ),
            ),
            build_scheduler=dict(
                name="ExponentialLR",
                gamma=0.9,
            ),
            save_model=dict(),
            save_task=dict(),
            train=dict(
                total_steps=15000,
                log_step=100,
                eval_step=500,
                save_step=500,
                gradient_clipping=1.0,
                gradient_accumulate=1,
                valid_metric="event_onset_200ms_fms",
                valid_higher_better=True,
                auto_resume=True,
                resume_ckpt_dir=None,
            ),
            evaluate=dict(),
        )

    def prepare_data(
        self,
        prepare_data: dict,
        target_dir: str,
        cache_dir: str,
        get_path_only: bool = False,
    ):
        return dcase_2016_task2(
            **self._get_current_arguments(flatten_dict="prepare_data")
        )

    def build_dataset(
        self,
        build_dataset: dict,
        target_dir: str,
        cache_dir: str,
        mode: str,
        data_csv: str,
        encoder_path: str,
        frame_shift: int,
    ):
        @dataclass
        class Config:
            train: dict = None
            valid: dict = None
            test: dict = None

        conf = Config(**build_dataset)
        conf = getattr(conf, mode)
        conf = conf or {}

        with open(encoder_path, "rb") as f:
            encoder = pickle.load(f)

        df = pd.read_csv(data_csv)
        df["label"] = [encoder.encode(label) for label in df["labels"].tolist()]

        dataset = FrameLabelDataset(df, len(encoder), frame_shift, **conf)
        return dataset

    def build_batch_sampler(
        self,
        build_batch_sampler: dict,
        target_dir: str,
        cache_dir: str,
        mode: str,
        data_csv: str,
        dataset,
    ):
        @dataclass
        class Config:
            train: dict = None
            valid: dict = None
            test: dict = None

        conf = Config(**build_batch_sampler)
        if mode == "train":
            return FixedBatchSizeBatchSampler(dataset, **(conf.train or {}))
        elif mode == "valid":
            record_ids = get_info(dataset, ["record_id"], target_dir / "valid_stats")
            return GroupSameItemSampler(record_ids)
        elif mode == "test":
            record_ids = get_info(dataset, ["record_id"], target_dir / "test_stats")
            return GroupSameItemSampler(record_ids)
        else:
            raise ValueError(f"Unsupported mode: {mode}")

    def build_task(
        self,
        build_task: dict,
        model: torch.nn.Module,
        encoder,
        valid_df: pd.DataFrame = None,
        test_df: pd.DataFrame = None,
    ):
        def df_to_events(df: pd.DataFrame):
            data = {}
            for rowid, row in df.iterrows():
                record_id = row["record_id"]
                if not record_id in data:
                    data[record_id] = []
                data[record_id].append(
                    {
                        "start": row["start_sec"] * 1000,
                        "end": row["end_sec"] * 1000,
                        "label": row["labels"],
                    }
                )
            return data

        valid_events = None if valid_df is None else df_to_events(valid_df)
        test_events = None if test_df is None else df_to_events(test_df)

        return EventPredictionTask(
            model,
            encoder,
            valid_target_events=valid_events,
            test_target_events=test_events,
            **build_task,
        )



import logging

import torch
from transformers import HubertModel, Wav2Vec2FeatureExtractor

SAMPLE_RATE = 16000
EXAMPLE_SEC = 5

logger = logging.getLogger(__name__)


class UpstreamExpert(torch.nn.Module):
    def __init__(self, ckpt, **kwds):
        super().__init__()
        try:
            self.extracter = Wav2Vec2FeatureExtractor.from_pretrained(ckpt)
        except:
            alter_extractor = "facebook/hubert-base-ls960"
            logger.info(
                f"The model {ckpt} on huggingface does not have a correspoinding feature extractor. "
                f"Using {alter_extractor}'s feature extractor as the alternative."
            )
            self.extracter = Wav2Vec2FeatureExtractor.from_pretrained(alter_extractor)
        self.model = HubertModel.from_pretrained(ckpt)

    def get_downsample_rates(self, key: str = None) -> int:
        return 320

    def forward(self, wavs):
        device = wavs[0].device
        wavs = [wav.detach().cpu().numpy() for wav in wavs]
        input_values = self.extracter(
            wavs,
            return_tensors="pt",
            padding=True,
            return_attention_mask=True,
            sampling_rate=SAMPLE_RATE,
        ).to(device)
        output_values = self.model(**input_values, output_hidden_states=True)

        return {"hidden_states": output_values.hidden_states}



from dataclasses import dataclass

from .base import AugmentedDynamicItemDataset, DataPipe


@dataclass
class NormWavDecibel(DataPipe):
    target_level: int = -25
    wav_name: str = "wav"
    norm_wav_name: str = "wav"
    """
    Args:
        target_level (int): normalize the wav decibel level to the target value
        wav_name (str): handle for the `takes` (input)
        norm_wav_name (str): handle for the `provides` (output)
    """

    def normalize_wav_decibel(self, wav):
        wav = wav.squeeze()  # (seq_len, 1) -> (seq_len,)
        if self.target_level == 0:
            return wav
        rms = wav.pow(2).mean().pow(0.5)
        scalar = (10 ** (self.target_level / 20)) / (rms + 1e-10)
        wav = wav * scalar
        return wav.unsqueeze(1)  # (seq_len,) -> (seq_len, 1)

    def __call__(self, dataset: AugmentedDynamicItemDataset):
        dataset.add_dynamic_item(
            self.normalize_wav_decibel, takes=self.wav_name, provides=self.norm_wav_name
        )
        return dataset



"""Resnet-ish adaptation for Pytorch

Implementation largely based on https://github.com/daisukelab/sound-clf-pytorch

Original paper:
@inproceedings{hershey2017cnn,
  title={CNN architectures for large-scale audio classification},
  author={Hershey, Shawn and Chaudhuri, Sourish and Ellis, Daniel PW and Gemmeke, Jort F and Jansen, Aren and Moore, R Channing and Plakal, Manoj and Platt, Devin and Saurous, Rif A and Seybold, Bryan and others},
  booktitle={2017 ieee international conference on acoustics, speech and signal processing (icassp)},
  pages={131--135},
  year={2017},
  organization={IEEE}
}
"""

from typing import Any, Callable, List, Optional, Type, Union

import torch
import torch.nn as nn
from torch import Tensor
from torch.nn.utils import weight_norm


def conv3x3(
    in_planes: int,
    out_planes: int,
    stride: int = 1,
    groups: int = 1,
    dilation: int = 1,
    standardize_weights: bool = False,
) -> nn.Conv2d:
    """3x3 convolution with padding"""
    conv = nn.Conv2d(
        in_planes,
        out_planes,
        kernel_size=3,
        stride=stride,
        padding=dilation,
        groups=groups,
        bias=False,
        dilation=dilation,
    )
    return weight_norm(conv) if standardize_weights else conv


def conv1x1(
    in_planes: int, out_planes: int, stride: int = 1, standardize_weights: bool = False
) -> nn.Conv2d:
    """1x1 convolution"""
    conv = nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)
    return weight_norm(conv) if standardize_weights else conv


class BasicBlock(nn.Module):
    expansion: int = 1

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
        standardize_weights: bool = False,
    ) -> None:
        super(BasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError("BasicBlock only supports groups=1 and base_width=64")
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv3x3(
            inplanes, planes, stride, standardize_weights=standardize_weights
        )
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes, standardize_weights=standardize_weights)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    # Bottleneck in torchvision places the stride for downsampling at 3x3 convolution(self.conv2)
    # while original implementation places the stride at the first 1x1 convolution(self.conv1)
    # according to "Deep residual learning for image recognition"https://arxiv.org/abs/1512.03385.
    # This variant is also known as ResNet V1.5 and improves accuracy according to
    # https://ngc.nvidia.com/catalog/model-scripts/nvidia:resnet_50_v1_5_for_pytorch.

    expansion: int = 4

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
        standardize_weights: bool = False,
    ) -> None:
        super(Bottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * (base_width / 64.0)) * groups
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv1x1(inplanes, width, standardize_weights=standardize_weights)
        self.bn1 = norm_layer(width)
        self.conv2 = conv3x3(
            width,
            width,
            stride,
            groups,
            dilation,
            standardize_weights=standardize_weights,
        )
        self.bn2 = norm_layer(width)
        self.conv3 = conv1x1(
            width, planes * self.expansion, standardize_weights=standardize_weights
        )
        self.bn3 = norm_layer(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x: Tensor) -> Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class ResNetish(nn.Module):
    # sample rate and embedding sizes are required model attributes for the HEAR API
    sample_rate = 16000
    embedding_size = 2048
    scene_embedding_size = embedding_size
    timestamp_embedding_size = embedding_size

    def __init__(
        self,
        block: Type[Union[BasicBlock, Bottleneck]],
        layers: List[int],
        num_classes: int = 1000,
        zero_init_residual: bool = False,
        groups: int = 1,
        width_per_group: int = 64,
        replace_stride_with_dilation: Optional[List[bool]] = None,
        norm_layer: Optional[Callable[..., nn.Module]] = None,
        standardize_weights: bool = False,
    ) -> None:
        super(ResNetish, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        self._norm_layer = norm_layer

        self.inplanes = 64
        self.dilation = 1
        if replace_stride_with_dilation is None:
            # each element in the tuple indicates if we should replace
            # the 2x2 stride with a dilated convolution instead
            replace_stride_with_dilation = [False, False, False]
        if len(replace_stride_with_dilation) != 3:
            raise ValueError(
                "replace_stride_with_dilation should be None "
                "or a 3-element tuple, got {}".format(replace_stride_with_dilation)
            )
        self.groups = groups
        self.base_width = width_per_group
        conv1 = nn.Conv2d(
            1, self.inplanes, kernel_size=7, stride=1, padding=3, bias=False
        )
        self.conv1 = weight_norm(conv1) if standardize_weights else conv1
        # self.conv1 = nn.Conv2d(1, self.inplanes, kernel_size=7, stride=1, padding=3, bias=False)
        self.bn1 = norm_layer(self.inplanes)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(
            block, 64, layers[0], standardize_weights=standardize_weights
        )
        self.layer2 = self._make_layer(
            block,
            128,
            layers[1],
            stride=2,
            dilate=replace_stride_with_dilation[0],
            standardize_weights=standardize_weights,
        )
        self.layer3 = self._make_layer(
            block,
            256,
            layers[2],
            stride=2,
            dilate=replace_stride_with_dilation[1],
            standardize_weights=standardize_weights,
        )
        self.layer4 = self._make_layer(
            block,
            512,
            layers[3],
            stride=2,
            dilate=replace_stride_with_dilation[2],
            standardize_weights=standardize_weights,
        )
        # self.avgpool = nn.AvgPool2d((4, 6))
        # self.fc = nn.Linear(512 * 24 * block.expansion, num_classes)
        # self.flatten = nn.Flatten()
        # self.add_max_mean = Lambda(lambda x: x.mean(1) + x.amax(1))

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)  # type: ignore[arg-type]
                elif isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)  # type: ignore[arg-type]

    def _make_layer(
        self,
        block: Type[Union[BasicBlock, Bottleneck]],
        planes: int,
        blocks: int,
        stride: int = 1,
        dilate: bool = False,
        standardize_weights: bool = False,
    ) -> nn.Sequential:
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(
                    self.inplanes, planes * block.expansion, stride, standardize_weights
                ),
                norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(
            block(
                self.inplanes,
                planes,
                stride,
                downsample,
                self.groups,
                self.base_width,
                previous_dilation,
                norm_layer,
            )
        )
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(
                block(
                    self.inplanes,
                    planes,
                    groups=self.groups,
                    base_width=self.base_width,
                    dilation=self.dilation,
                    norm_layer=norm_layer,
                )
            )

        return nn.Sequential(*layers)

    def _forward_impl(self, x: Tensor) -> Tensor:
        # See note [TorchScript super()]
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        # BYOL-A max-mean operation
        x = x.permute(0, 3, 2, 1)  # (batch, time, mel, ch)
        B, T, D, C = x.shape
        x = x.reshape((B, T, C * D))  # (batch, time, mel*ch)

        x = x.mean(1) + x.amax(1)

        return x

    def forward(self, x: Tensor) -> Tensor:
        return self._forward_impl(x)


def _resnetish(
    arch: str,
    block: Type[Union[BasicBlock, Bottleneck]],
    layers: List[int],
    pretrained: bool,
    progress: bool,
    **kwargs: Any
) -> ResNetish:
    model = ResNetish(block, layers, **kwargs)
    # if pretrained:
    # state_dict = torch.utils.load_state_dict_from_url(model_urls[arch], progress=progress)
    # model.load_state_dict(state_dict)
    return model


def resnetish10(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> ResNetish:
    r"""ResNet-10 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnetish(
        "resnetish18", BasicBlock, [1, 1, 1, 1], pretrained, progress, **kwargs
    )


def resnetish18(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> ResNetish:
    r"""ResNet-18 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnetish(
        "resnetish18", BasicBlock, [2, 2, 2, 2], pretrained, progress, **kwargs
    )


def resnetish34(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> ResNetish:
    r"""ResNet-34 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_.
    Adapted for Audio from
    `"CNN architectures for large-scale audio classification" <https://arxiv.org/abs/1609.09430>`_.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnetish(
        "resnetish34", BasicBlock, [3, 4, 6, 3], pretrained, progress, **kwargs
    )


def resnetish50(
    pretrained: bool = False, progress: bool = True, **kwargs: Any
) -> ResNetish:
    r"""ResNet-50 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_.
    Adapted for Audio from
    `"CNN architectures for large-scale audio classification" <https://arxiv.org/abs/1609.09430>`_.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    return _resnetish(
        "resnet50", Bottleneck, [3, 4, 6, 3], pretrained, progress, **kwargs
    )



from .base import Dataset
from .diarization import DiarizationDataset
from .encode import EncodeCategories, EncodeCategory, EncodeMultiLabel, EncodeText
from .frame_label import FrameLabelDataset
from .load_audio import LoadAudio
from .util import get_info



"""
Utilities for file format conversion for Speaker Diarization

Authors:
  * Jiatong Shi 2021
  * Leo 2022
"""

import logging
import os
import re
import subprocess
from pathlib import Path
from typing import List

import numpy as np
import pandas as pd
import torch
from scipy.signal import medfilt
from tqdm import tqdm

logger = logging.getLogger(__name__)

RTTM_FORMAT = "SPEAKER {:s} 1 {:7.2f} {:7.2f} <NA> <NA> {:s} <NA>"

__all__ = [
    "kaldi_dir_to_rttm",
    "csv_to_kaldi_dir",
    "kaldi_dir_to_csv",
]


def kaldi_dir_to_rttm(data_dir: str, rttm_path: str):
    data_dir: Path = Path(data_dir)
    segments_file = data_dir / "segments"
    utt2spk_file = data_dir / "utt2spk"

    assert segments_file.is_file()
    assert utt2spk_file.is_file()

    utt2spk = {}
    with utt2spk_file.open() as f:
        for utt2spk_line in f.readlines():
            fields = utt2spk_line.strip().replace("\n", " ").split()
            assert len(fields) == 2
            utt, spk = fields
            utt2spk[utt] = spk

    with Path(rttm_path).open("w") as rttm_f:
        with segments_file.open() as f:
            for segment_line in f.readlines():
                fields = segment_line.strip().replace("\t", " ").split()
                assert len(fields) == 4
                utt, reco, start, end = fields
                spk = utt2spk[utt]
                print(
                    RTTM_FORMAT.format(
                        reco,
                        float(start),
                        float(end) - float(start),
                        spk,
                    ),
                    file=rttm_f,
                )


def make_rttm_and_score(
    prediction_dir: str,
    score_dir: str,
    gt_rttm: str,
    frame_shift: int,
    thresholds: List[int],
    medians: List[int],
    subsampling: int = 1,
    sampling_rate: int = 16000,
):
    Path(score_dir).mkdir(exist_ok=True, parents=True)
    dscore_dir = Path(score_dir) / "dscore"
    rttm_dir = Path(score_dir) / "rttm"
    result_dir = Path(score_dir) / "result"

    setting2dscore = []
    for th in thresholds:
        for med in medians:
            logger.info(f"Make RTTM with threshold {th}, median filter {med}")
            rttm_file = rttm_dir / f"threshold-{th}_median-{med}.rttm"
            make_rttm(
                prediction_dir,
                rttm_file,
                th,
                med,
                frame_shift,
                subsampling,
                sampling_rate,
            )

            logger.info(f"Scoring...")
            result_file = result_dir / f"threshold-{th}_median-{med}.result"
            overall_der = score_with_dscore(dscore_dir, rttm_file, gt_rttm, result_file)
            logger.info(f"DER: {overall_der}")

            setting2dscore.append(((th, med), overall_der))

    setting2dscore.sort(key=lambda x: x[1])
    (best_th, best_med), best_der = setting2dscore[0]
    return best_der, (best_th, best_med)


def make_rttm(
    prediction_dir: str,
    out_rttm_path: str,
    threshold: int,
    median: int,
    frame_shift: int,
    subsampling: int,
    sampling_rate: int,
):
    names = sorted([name for name in os.listdir(prediction_dir)])
    filepaths = [Path(prediction_dir) / name for name in names]

    Path(out_rttm_path).parent.mkdir(exist_ok=True, parents=True)
    with open(out_rttm_path, "w") as wf:
        for filepath in filepaths:
            session, _ = os.path.splitext(os.path.basename(filepath))
            data = torch.load(filepath).numpy()
            a = np.where(data > threshold, 1, 0)
            if median > 1:
                a = medfilt(a, (median, 1))
            factor = frame_shift * subsampling / sampling_rate
            for spkid, frames in enumerate(a.T):
                frames = np.pad(frames, (1, 1), "constant")
                (changes,) = np.where(np.diff(frames, axis=0) != 0)
                for s, e in zip(changes[::2], changes[1::2]):
                    print(
                        RTTM_FORMAT.format(
                            session,
                            s * factor,
                            (e - s) * factor,
                            session + "_" + str(spkid),
                        ),
                        file=wf,
                    )


def score_with_dscore(
    dscore_dir: str, hyp_rttm: str, gt_rttm: str, score_result: str
) -> float:
    """
    This function returns the overall DER score, and will also write the detailed scoring results
    to 'score_result'
    """
    dscore_dir: Path = Path(dscore_dir)
    Path(score_result).parent.mkdir(exist_ok=True, parents=True)

    if not dscore_dir.is_dir():
        logger.info(f"Cloning dscore into {dscore_dir}")
        subprocess.check_output(
            f"git clone https://github.com/nryant/dscore.git {dscore_dir}",
            shell=True,
        ).decode("utf-8")

    subprocess.check_call(
        f"python3 {dscore_dir}/score.py -r {gt_rttm} -s {hyp_rttm} > {score_result}",
        shell=True,
    )

    return get_overall_der_from_dscore_file(score_result)


def get_overall_der_from_dscore_file(score_result: str):
    with open(score_result) as file:
        lines = file.readlines()
        overall_lines = [line for line in lines if "OVERALL" in line]
        assert len(overall_lines) == 1
        overall_line = overall_lines[0]
        overall_line = re.sub("\t+", " ", overall_line)
        overall_line = re.sub(" +", " ", overall_line)
        overall_der = float(overall_line.split(" ")[3])
        # The overall der line should look like:
        # *** OVERALL *** DER JER ...
    return overall_der


def csv_to_kaldi_dir(csv: str, data_dir: str):
    logger.info(f"Convert csv {csv} into kaldi data directory {data_dir}")

    data_dir: Path = Path(data_dir)
    data_dir.mkdir(exist_ok=True, parents=True)

    df = pd.read_csv(csv)
    required = ["record_id", "wav_path", "utt_id", "speaker", "start_sec", "end_sec"]
    for r in required:
        assert r in df.columns

    reco2path = {}
    reco2dur = {}
    utt2spk = {}
    spk2utt = {}
    segments = []
    for rowid, row in tqdm(df.iterrows(), total=len(df)):
        record_id, wav_path, duration, utt_id, speaker, start_sec, end_sec = (
            row["record_id"],
            row["wav_path"],
            row["duration"],
            row["utt_id"],
            row["speaker"],
            row["start_sec"],
            row["end_sec"],
        )
        if record_id in reco2path:
            assert wav_path == reco2path[record_id]
        else:
            reco2path[record_id] = wav_path

        if record_id not in reco2dur:
            reco2dur[record_id] = duration
        else:
            assert reco2dur[record_id] == duration

        if utt_id not in utt2spk:
            utt2spk[utt_id] = str(speaker)
        else:
            assert utt2spk[utt_id] == str(speaker)

        if speaker not in spk2utt:
            spk2utt[speaker] = []
        spk2utt[speaker].append(utt_id)

        segments.append((utt_id, record_id, str(start_sec), str(end_sec)))

    with (data_dir / "wav.scp").open("w") as f:
        f.writelines([f"{reco} {path}\n" for reco, path in reco2path.items()])

    with (data_dir / "reco2dur").open("w") as f:
        f.writelines([f"{reco} {dur}\n" for reco, dur in reco2dur.items()])

    with (data_dir / "utt2spk").open("w") as f:
        f.writelines([f"{utt} {spk}\n" for utt, spk in utt2spk.items()])

    with (data_dir / "spk2utt").open("w") as f:
        f.writelines([f"{spk} {' '.join(utts)}\n" for spk, utts in spk2utt.items()])

    with (data_dir / "segments").open("w") as f:
        f.writelines(
            [f"{utt} {record} {start} {end}\n" for utt, record, start, end in segments]
        )


def kaldi_dir_to_csv(data_dir: str, csv: str):
    logger.info(f"Convert kaldi data directory {data_dir} into csv {csv}")

    data_dir: Path = Path(data_dir)

    assert (data_dir / "wav.scp").is_file()
    assert (data_dir / "segments").is_file()
    assert (data_dir / "utt2spk").is_file()
    assert (data_dir / "reco2dur").is_file()

    reco2path = {}
    with (data_dir / "wav.scp").open() as f:
        for line in f.readlines():
            line = line.strip()
            reco, path = line.split(" ")
            reco2path[reco] = path

    reco2dur = {}
    with (data_dir / "reco2dur").open() as f:
        for line in f.readlines():
            line = line.strip()
            reco, duration = line.split(" ")
            reco2dur[reco] = float(duration)

    utt2spk = {}
    with (data_dir / "utt2spk").open() as f:
        for line in f.readlines():
            line = line.strip()
            utt, spk = line.split(" ")
            utt2spk[utt] = spk

    row = []
    with (data_dir / "segments").open("r") as f:
        for line in f.readlines():
            line = line.strip()
            utt, reco, start, end = line.split(" ")
            row.append(
                (
                    reco,
                    reco2path[reco],
                    reco2dur[reco],
                    utt,
                    utt2spk[utt],
                    float(start),
                    float(end),
                )
            )

    recos, wav_paths, durations, utts, spks, starts, ends = zip(*row)
    pd.DataFrame(
        data=dict(
            record_id=recos,
            wav_path=wav_paths,
            utt_id=utts,
            speaker=spks,
            start_sec=starts,
            end_sec=ends,
            duration=durations,
        )
    ).to_csv(csv, index=False)



# -*- coding: utf-8 -*- #
"""*********************************************************************************************"""
#   FileName     [ upstream/cpc/cpc_default_config.py ]
#   Synopsis     [ the cpc default config ]
#   Author       [ Facebook Research ]
#   Copyright    [ Facebook, Inc. and its affiliates ]
#   Reference    [ https://github.com/facebookresearch/CPC_audio ]
"""*********************************************************************************************"""


###############
# IMPORTATION #
###############
import argparse


def get_default_cpc_config():
    parser = set_default_cpc_config(argparse.ArgumentParser())
    return parser.parse_args([])


def set_default_cpc_config(parser):
    # Run parameters

    group = parser.add_argument_group(
        "Architecture configuration",
        description="The arguments defining the " "model's architecture.",
    )
    group.add_argument(
        "--hiddenEncoder",
        type=int,
        default=256,
        help="Hidden dimension of the encoder network.",
    )
    group.add_argument(
        "--hiddenGar",
        type=int,
        default=256,
        help="Hidden dimension of the auto-regressive network",
    )
    group.add_argument(
        "--nPredicts", type=int, default=12, help="Number of steps to predict."
    )
    group.add_argument(
        "--negativeSamplingExt",
        type=int,
        default=128,
        help="Number of negative samples to take.",
    )
    group.add_argument("--learningRate", type=float, default=2e-4)
    group.add_argument(
        "--schedulerStep",
        type=int,
        default=-1,
        help="Step of the learning rate scheduler: at each "
        "step the learning rate is divided by 2. Default: "
        "no scheduler.",
    )
    group.add_argument(
        "--schedulerRamp",
        type=int,
        default=None,
        help="Enable a warm up phase for the learning rate: "
        "adds a linear ramp of the given size.",
    )
    group.add_argument(
        "--beta1", type=float, default=0.9, help="Value of beta1 for the Adam optimizer"
    )
    group.add_argument(
        "--beta2",
        type=float,
        default=0.999,
        help="Value of beta2 for the Adam optimizer",
    )
    group.add_argument(
        "--epsilon",
        type=float,
        default=1e-08,
        help="Value of epsilon for the Adam optimizer",
    )
    group.add_argument(
        "--sizeWindow",
        type=int,
        default=20480,
        help="Number of frames to consider at each batch.",
    )
    group.add_argument("--nEpoch", type=int, default=200, help="Number of epoch to run")
    group.add_argument(
        "--samplingType",
        type=str,
        default="samespeaker",
        choices=["samespeaker", "uniform", "samesequence", "sequential"],
        help="How to sample the negative examples in the " "CPC loss.",
    )
    group.add_argument(
        "--nLevelsPhone",
        type=int,
        default=1,
        help="(Supervised mode only). Number of layers in "
        "the phone classification network.",
    )
    group.add_argument(
        "--cpc_mode",
        type=str,
        default=None,
        choices=["reverse", "none"],
        help="Some variations on CPC.",
    )
    group.add_argument(
        "--encoder_type",
        type=str,
        choices=["cpc", "mfcc", "lfb"],
        default="cpc",
        help="Replace the encoder network by mfcc features " "or learned filter banks",
    )
    group.add_argument(
        "--normMode",
        type=str,
        default="layerNorm",
        choices=["instanceNorm", "ID", "layerNorm", "batchNorm"],
        help="Type of normalization to use in the encoder "
        "network (default is layerNorm).",
    )
    group.add_argument(
        "--onEncoder",
        action="store_true",
        help="(Supervised mode only) Perform the "
        "classification on the encoder's output.",
    )
    group.add_argument(
        "--random_seed", type=int, default=None, help="Set a specific random seed."
    )
    group.add_argument(
        "--speakerEmbedding",
        type=int,
        default=0,
        help="(Depreciated) Feed the prediction network with "
        "speaker embeddings along with the usual sequence.",
    )
    group.add_argument(
        "--arMode",
        default="LSTM",
        choices=["GRU", "LSTM", "RNN", "no_ar", "transformer"],
        help="Architecture to use for the auto-regressive "
        "network (default is lstm).",
    )
    group.add_argument(
        "--nLevelsGRU",
        type=int,
        default=1,
        help="Number of layers in the autoregressive network.",
    )
    group.add_argument(
        "--rnnMode",
        type=str,
        default="transformer",
        choices=[
            "transformer",
            "RNN",
            "LSTM",
            "linear",
            "ffd",
            "conv4",
            "conv8",
            "conv12",
        ],
        help="Architecture to use for the prediction network",
    )
    group.add_argument(
        "--dropout",
        action="store_true",
        help="Add a dropout layer at the output of the " "prediction network.",
    )
    group.add_argument(
        "--abspos",
        action="store_true",
        help="If the prediction network is a transformer, "
        "active to use absolute coordinates.",
    )

    return parser
